"""Global settings for the package."""

from __future__ import annotations

from enum import Enum

from sage_acsv.debug import acsv_logger
from sage.all import Macaulay2


class OutputFormat(Enum):
    """Output options for displaying the asymptotic behavior determined
    by :func:`.diagonal_asy`.

    See also:

    - :func:`.diagonal_asy`
    - :class:`.ACSVSettings`

    """

    ASYMPTOTIC = "asymptotic"
    SYMBOLIC = "symbolic"
    TUPLE = "tuple"


class KroneckerBackend(Enum):
    """Options for computing Kronecker representations determined by
    :func:`.ContributingCombinatorial`, :func:`.MinimalCriticalCombinatorial`, and
    :func:`.CriticalPoints`.

    See also:

    - :func:`.ContributingCombinatorial`
    - :func:`.MinimalCriticalCombinatorial`
    - :func:`.CriticalPoints`
    - :class:`.ACSVSettings`

    """

    SAGE_ACSV = "sage_acsv"
    MSOLVE = "msolve"


class GroebnerBackend(Enum):
    """Options for computing Groebner Bases and related ideal functions
    :func:`.GroebnerBasis`, :func:`.PrimaryDecomposition`, and :func:`.CriticalPoints`.

    See also:

    - :func:`.GroebnerBasis`
    - :func:`.PrimaryDecomposition`
    - :func:`.Radical`
    - :class:`.ACSVSettings`

    """

    SINGULAR = "singular"
    MACAULAY2 = "macaulay2"


DEFAULT_OUTPUT_FORMAT = OutputFormat.ASYMPTOTIC
DEFAULT_KRONECKER_BACKEND = KroneckerBackend.SAGE_ACSV
DEFAULT_GROEBNER_BACKEND = GroebnerBackend.SINGULAR


class ACSVSettings:
    """Global settings for the package.

    This class is used to store global settings for the package. It
    provides a mechanism to set and get the default settings.

    Examples::

        sage: from sage_acsv import ACSVSettings
        sage: list(ACSVSettings.Output)
        [<OutputFormat.ASYMPTOTIC: 'asymptotic'>,
         <OutputFormat.SYMBOLIC: 'symbolic'>,
         <OutputFormat.TUPLE: 'tuple'>]
        sage: ACSVSettings.Output.ASYMPTOTIC
        <OutputFormat.ASYMPTOTIC: 'asymptotic'>
        sage: ACSVSettings.set_default_output_format("tuple")
        sage: ACSVSettings.get_default_output_format()
        <OutputFormat.TUPLE: 'tuple'>
        sage: ACSVSettings.set_default_output_format("symbolic")
        sage: ACSVSettings.get_default_output_format()
        <OutputFormat.SYMBOLIC: 'symbolic'>
        sage: ACSVSettings.set_default_output_format()
        sage: ACSVSettings.get_default_output_format()
        <OutputFormat.ASYMPTOTIC: 'asymptotic'>

    """

    Output = OutputFormat
    Kronecker = KroneckerBackend
    Groebner = GroebnerBackend
    _default_output_format = DEFAULT_OUTPUT_FORMAT
    _default_kronecker_backend = DEFAULT_KRONECKER_BACKEND
    _default_groebner_backend = DEFAULT_GROEBNER_BACKEND
    _m2 = Macaulay2()

    MAX_MIN_CRIT_RETRIES = 5  # Maximum number of retries for critical point detection

    @classmethod
    def set_default_output_format(
        cls, output_format: OutputFormat | str | None = None
    ) -> None:
        """Set the default output format chosen when not explicitly
        specifying one.

        INPUT:

        * ``output_format`` -- a given :class:`.OutputFormat`, a string identifying one,
          or ``None`` to restore the default behavior.
        """
        if output_format is None:
            cls._default_output_format = DEFAULT_OUTPUT_FORMAT
        else:
            cls._default_output_format = OutputFormat(output_format)

    @classmethod
    def get_default_output_format(cls) -> OutputFormat:
        """Get the default output format chosen when not explicitly
        specifying one.
        """
        return cls._default_output_format

    @classmethod
    def get_logging_level(cls) -> int:
        """Get the logging level for the package."""
        return acsv_logger.level

    @classmethod
    def set_logging_level(cls, level: int) -> None:
        """Set the logging level for the package.

        INPUT:

        * ``level`` -- a logging level.

        EXAMPLES::

            sage: from sage_acsv import ACSVSettings
            sage: import logging
            sage: ACSVSettings.set_logging_level(logging.DEBUG)
            sage: ACSVSettings.get_logging_level()
            10
            sage: ACSVSettings.set_logging_level(logging.INFO)
            sage: ACSVSettings.get_logging_level()
            20
        """
        acsv_logger.setLevel(level)

    @classmethod
    def set_default_kronecker_backend(
        cls, backend: KroneckerBackend | str | None
    ) -> None:
        """Set the preferred method for computing a Kronecker Representation. Will default
        to the sage_acsv implementation if not specified.

        See :func:`._kronecker_representation`.

        INPUT:

        * ``backend`` -- a given :class:`.KroneckerBackend`, a string identifying one,
          or ``None`` to restore the default behavior.
        """
        if backend is None:
            cls._default_kronecker_backend = DEFAULT_KRONECKER_BACKEND
        else:
            cls._default_kronecker_backend = KroneckerBackend(backend)

    @classmethod
    def get_default_kronecker_backend(cls) -> KroneckerBackend:
        """Get the default kronecker representation backend."""
        return cls._default_kronecker_backend

    @classmethod
    def set_default_groebner_backend(
        cls, backend: GroebnerBackend | str | None
    ) -> None:
        """Set the preferred method for performing Groebner Bases and related computations,
        including :func:`.GroebnerBasis`, :func:`.PrimaryDecomposition`, and :func:`.Radical`

        Will default to singular.

        INPUT:

        * ``backend`` -- a given :class:`.GroebnerBackend`, a string identifying one,
          or ``None`` to restore the default behavior.
        """
        if backend is None:
            cls._default_groebner_backend = DEFAULT_GROEBNER_BACKEND
        else:
            cls._default_groebner_backend = GroebnerBackend(backend)

    @classmethod
    def get_default_groebner_backend(cls) -> KroneckerBackend:
        """Get the default groebner basis backend."""
        return cls._default_groebner_backend

    @classmethod
    def set_macaulay2_path(cls, path: str | None) -> None:
        """Set the path of the Macaulay2 executable.

        INPUT:

        * ``path`` -- a filepath string or `None` to run from anywhere
        """
        cls._m2 = Macaulay2(command=path)

    @classmethod
    def get_macaulay2(cls) -> KroneckerBackend:
        """Get the Macaulay2 installation path."""
        return cls._m2



import os
import tempfile
import subprocess

from sage.misc.sage_eval import sage_eval
from sage.features.msolve import msolve
from sage_acsv.helpers import ACSVException


def get_parametrization(vs, system):
    filename = msolve().absolute_filename()
    msolve_in = tempfile.NamedTemporaryFile(mode="w", encoding="ascii", delete=False)
    command = [filename, "-f", msolve_in.name, "-P", "2"]

    system = list(str(e) for e in system)
    try:
        print(",".join([str(v) for v in vs]), file=msolve_in)
        print(0, file=msolve_in)
        print(*(pol.replace(" ", "") for pol in system), sep=",\n", file=msolve_in)
        msolve_in.close()
        msolve_out = subprocess.run(command, capture_output=True, text=True)
    finally:
        os.unlink(msolve_in.name)

    msolve_out.check_returncode()

    result = msolve_out.stdout
    result = sage_eval(result[:-2])

    if result[0] != 0:
        raise ACSVException(
            "Issue with msolve parametrization - system does not have finitely many solutions"
        )

    return result



"""Functions for determining asymptotics of the coefficients
of multivariate rational functions.
"""

from copy import copy

from sage.all import AA, PolynomialRing, QQ, QQbar, SR, DifferentialWeylAlgebra, Ideal
from sage.all import (
    gcd,
    prod,
    pi,
    matrix,
    exp,
    log,
    I,
    factorial,
    srange,
    shuffle,
    vector,
)

from sage_acsv.kronecker import _kronecker_representation
from sage_acsv.helpers import (
    ACSVException,
    IsContributing,
    NewtonSeries,
    RationalFunctionReduce,
    GetHessian,
    ImplicitHessian,
    collapse_zero_part,
)
from sage_acsv.debug import Timer, acsv_logger
from sage_acsv.settings import ACSVSettings
from sage_acsv.whitney import WhitneyStrat
from sage_acsv.macaulay2 import PrimaryDecomposition, Saturate


# we need to monkeypatch a function from the asymptotics module such that creating
# asymptotic expansions over QQbar is possible. this should be removed once the
# upstream issue is resolved.

import sage.rings.asymptotic.misc as asy_misc

from sage.rings.integer_ring import ZZ

strip_symbolic_original = asy_misc.strip_symbolic


def strip_symbolic(expression):
    expression = strip_symbolic_original(expression)
    if expression in ZZ:
        expression = ZZ(expression)
    return expression


asy_misc.strip_symbolic = strip_symbolic


def _diagonal_asy_smooth(
    G,
    H,
    r=None,
    linear_form=None,
    expansion_precision=1,
    return_points=False,
    output_format=None,
    as_symbolic=False,
):
    r"""Asymptotics in a given direction `r` of the multivariate rational
    function `F = G/H` when the singular variety of `F` is smooth.

    The function is assumed to have a combinatorial expansion.

    INPUT:

    * ``G`` -- The numerator of the rational function ``F``.
    * ``H`` -- The denominator of the rational function ``F``.
    * ``r`` -- A vector of positive algebraic numbers (generally integers),
      one entry per variable of `F`. Defaults to the appropriate vector of
      all 1's if not specified.
    * ``linear_form`` -- (Optional) A linear combination of the input
      variables that separates the critical point solutions.
    * ``expansion_precision`` -- A positive integer value. This is the number
      of terms to compute in the asymptotic expansion. Defaults to 1, which
      only computes the leading term.
    * ``return_points`` -- If ``True``, also returns the coordinates of
      minimal critical points. By default ``False``.
    * ``output_format`` -- (Optional) A string or :class:`.ACSVSettings.Output`
      specifying the way the asymptotic growth is returned. Allowed values
      currently are:
      - ``"tuple"``: the growth is returned as a list of
        tuples of the form ``(a, n^b, pi^c, d)`` such that the `r`-diagonal of `F`
        is the sum of ``a^n n^b pi^c d + O(a^n n^{b-1})`` over these tuples.
      - ``"symbolic"``: the growth is returned as an expression from the symbolic
        ring ``SR`` in the variable ``n``.
      - ``"asymptotic"``: the growth is returned as an expression from an appropriate
        ``AsymptoticRing`` in the variable ``n``.
      - ``None``: the default, which uses the default set for
        :class:`.ACSVSettings.Output` itself via
        :meth:`.ACSVSettings.set_default_output_format`. The default behavior
        is asymptotic output.
    * ``as_symbolic`` -- deprecated in favor of the equivalent
      ``output_format="symbolic"``. Will be removed in a future release.

    OUTPUT:

    A representation of the asymptotic behavior of the coefficient sequence,
    either as a list of tuples, or as a symbolic expression.

    See also:

    - :func:`.diagonal_asy`

    TESTS:

    Check that passing a non-supported ``output_format`` errors out::

        sage: from sage_acsv import diagonal_asy
        sage: var('x y')
        (x, y)
        sage: diagonal_asy(1/(1 - x - y), output_format='hello world')  # indirect doctest
        Traceback (most recent call last):
        ...
        ValueError: 'hello world' is not a valid OutputFormat
        sage: diagonal_asy(1/(1 - x - y), output_format=42)  # indirect doctest
        Traceback (most recent call last):
        ...
        ValueError: 42 is not a valid OutputFormat

    """
    # Initialize variables
    vs = list(H.variables())

    t, lambda_, u_ = PolynomialRing(QQ, "t, lambda_, u_").gens()
    expanded_R = PolynomialRing(QQ, len(vs) + 3, vs + [t, lambda_, u_])

    vs = [expanded_R(v) for v in vs]
    t, lambda_, u_ = expanded_R(t), expanded_R(lambda_), expanded_R(u_)
    vsT = vs + [t, lambda_]

    d = len(vs)
    rd = r[-1]

    # Make sure G and H are coprime, and that H does not vanish at 0
    G, H = expanded_R(G), expanded_R(H)
    if H.subs({v: 0 for v in H.variables()}) == 0:
        raise ValueError("Denominator vanishes at 0.")

    # In case form doesn't separate, we want to try again
    for _ in range(ACSVSettings.MAX_MIN_CRIT_RETRIES):
        try:
            # Find minimal critical points in Kronecker Representation
            min_crit_pts = ContributingCombinatorialSmooth(
                G, H, vs, r=r, linear_form=linear_form
            )
            break
        except Exception as e:
            if isinstance(e, ACSVException) and e.retry:
                acsv_logger.info(
                    "Randomly generated linear form was not suitable, "
                    f"encountered error: {e}\nRetrying..."
                )
                continue
            else:
                raise e
    else:
        return

    timer = Timer()

    # Find det(zH_z Hess) where Hess is the Hessian of z_1...z_n * log(g(z_1, ..., z_n))
    Det = GetHessian(H, vsT[0:-2], r).determinant()

    # Find exponential growth
    T = prod([SR(vs[i]) ** r[i] for i in range(d)])

    # Find constants appearing in asymptotics in terms of original variables
    B = SR(1 / Det / rd ** (d - 1) / 2 ** (d - 1))
    C = SR(1 / T)

    # Compute constants at contributing singularities
    n = SR.var("n")
    asm_quantities = []
    for cp in min_crit_pts:
        subs_dict = {SR(v): V for (v, V) in zip(vs, cp)}
        expansion = sum(
            [
                term / (rd * n) ** (term_order)
                for term_order, term in enumerate(
                    GeneralTermAsymptotics(G, H, r, vs, cp, expansion_precision)
                )
            ]
        )
        B_sub = B.subs(subs_dict)
        C_sub = C.subs(subs_dict)
        try:
            B_sub = QQbar(B_sub)
            C_sub = QQbar(C_sub)
        except (ValueError, TypeError):
            pass
        asm_quantities.append([expansion, B_sub, C_sub])

    n = SR.var("n")
    asm_vals = [(c, QQ(1 - d) / 2, b.sqrt(), a) for (a, b, c) in asm_quantities]
    timer.checkpoint("Final Asymptotics")

    if as_symbolic:
        acsv_logger.warning(
            "The as_symbolic argument has been deprecated in favor of output_format='symbolic' "
        )
        output_format = ACSVSettings.Output.SYMBOLIC

    if output_format is None:
        output_format = ACSVSettings.get_default_output_format()
    else:
        output_format = ACSVSettings.Output(output_format)

    if output_format in (ACSVSettings.Output.TUPLE, ACSVSettings.Output.SYMBOLIC):
        n = SR.var("n")
        result = [
            (base, n**exponent, pi**exponent, constant * expansion)
            for (base, exponent, constant, expansion) in asm_vals
        ]
        if output_format == ACSVSettings.Output.SYMBOLIC:
            result = sum([a**n * b * c * d for (a, b, c, d) in result])

    elif output_format == ACSVSettings.Output.ASYMPTOTIC:
        from sage.all import AsymptoticRing

        AR = AsymptoticRing("QQbar^n * n^QQ", QQbar)
        n = AR.gen()
        result = sum(
            [  # bug in AsymptoticRing requires splitting out modulus manually
                constant
                * pi**exponent
                * abs(base) ** n
                * collapse_zero_part(base / abs(base)) ** n
                * n**exponent
                * AR(expansion)
                + (abs(base) ** n * n ** (exponent - expansion_precision)).O()
                for (base, exponent, constant, expansion) in asm_vals
            ]
        )

    else:
        raise NotImplementedError(f"Missing implementation for {output_format}")

    if return_points:
        return result, min_crit_pts

    return result


def diagonal_asy(
    F,
    r=None,
    linear_form=None,
    expansion_precision=1,
    return_points=False,
    output_format=None,
    whitney_strat=None,
    as_symbolic=False,
):
    r"""Asymptotic behavior of the coefficient array of a multivariate rational
    function `F` along a given direction `r`.

    INPUT:

    * ``F`` -- The rational function `G/H` in `d` variables. This function is
      assumed to have a combinatorial expansion.
    * ``r`` -- A vector of length `d` of positive algebraic numbers.
      Defaults to the appropriate vector of all 1's if not specified.
    * ``linear_form`` -- (Optional) A linear combination of the input
      variables that separates the critical point solutions. Is generated
      randomly if not specified.
    * ``expansion_precision`` -- A positive integer, the number of terms to
      compute in the asymptotic expansion. Defaults to 1, which only computes
      the leading term.
    * ``return_points`` -- If ``True``, also returns the coordinates of
      minimal critical points. By default ``False``.
    * ``output_format`` -- (Optional) A string or
      :class:`.ACSVSettings.Output` specifying the way the asymptotic growth
      is returned. Allowed values currently are:
      - ``"tuple"``: the growth is returned as a list of
        tuples of the form ``(a, n^b, pi^c, d)`` such that the `r`-diagonal of `F`
        behaves like the sum of ``a^n n^b pi^c d + O(a^n n^{b-1})`` over these tuples.
      - ``"symbolic"``: the growth is returned as an expression from the symbolic
        ring ``SR`` in the variable ``n``.
      - ``"asymptotic"``: the growth is returned as an expression from an appropriate
        ``AsymptoticRing`` in the variable ``n``.
      - ``None``: the default, which uses the default set for
        :class:`.ACSVSettings.Output` itself via
        :meth:`.ACSVSettings.set_default_output_format`. The default behavior
        is asymptotic output.
    * ``as_symbolic`` -- deprecated in favor of the equivalent
      ``output_format="symbolic"``. Will be removed in a future release.
    * ``whitney_strat`` -- (Optional) If known / precomputed, a
      Whitney Stratification of `V(H)`. The program will not check if
      this stratification is correct. Should be a list of length ``d``, where
      the ``k``-th entry is a list of tuples of ideas generators representing
      a component of the ``k``-dimensional stratum.

    OUTPUT:

    A representation of the asymptotic behavior of the coefficient array of `F` along
    the specified direction.

    NOTE:

    The code randomly generates a linear form, which for generic rational functions
    separates the solutions of an intermediate polynomial system with high probability.
    This separation step can fail, but (assuming `F` has a finite number of critical
    points) the code can be rerun until a separating form is found.

    EXAMPLES:

        sage: from sage_acsv import diagonal_asy
        sage: var('x,y,z,w')
        (x, y, z, w)
        sage: diagonal_asy(1/(1-x-y))
        1/sqrt(pi)*4^n*n^(-1/2) + O(4^n*n^(-3/2))
        sage: diagonal_asy(1/(1-(1+x)*y), r = [1,2], return_points=True)
        (1/sqrt(pi)*4^n*n^(-1/2) + O(4^n*n^(-3/2)), [[1, 1/2]])
        sage: diagonal_asy(1/(1-(x+y+z)+(3/4)*x*y*z), output_format="symbolic")
        0.840484893481498?*24.68093482214177?^n/(pi*n)
        sage: diagonal_asy(1/(1-(x+y+z)+(3/4)*x*y*z))
        0.840484893481498?/pi*24.68093482214177?^n*n^(-1) + O(24.68093482214177?^n*n^(-2))
        sage: var('n')
        n
        sage: asy = diagonal_asy(
        ....:     1/(1 - w*(1 + x)*(1 + y)*(1 + z)*(x*y*z + y*z + y + z + 1)),
        ....:     output_format="tuple",
        ....: )
        sage: sum([
        ....:      a.radical_expression()^n * b * c * QQbar(d).radical_expression()
        ....:      for (a, b, c, d) in asy
        ....: ])
        1/4*(12*sqrt(2) + 17)^n*sqrt(17/2*sqrt(2) + 12)/(pi^(3/2)*n^(3/2))

    Not specifying any ``output_format`` falls back to the default asymptotic
    representation::

        sage: diagonal_asy(1/(1 - 2*x))
        2^n + O(2^n*n^(-1))
        sage: diagonal_asy(1/(1 - 2*x), output_format="tuple")
        [(2, 1, 1, 1)]

    Passing ``"symbolic"`` lets the function return an element of the
    symbolic ring in the variable ``n`` that describes the asymptotic growth::

        sage: growth = diagonal_asy(1/(1 - 2*x), output_format="symbolic"); growth
        2^n
        sage: growth.parent()
        Symbolic Ring

    The argument ``"asymptotic"`` constructs an asymptotic expansion over
    an appropriate ``AsymptoticRing`` in the variable ``n``, including the
    appropriate error term::

        sage: assume(SR.an_element() > 0)  # required to make coercions involving SR work properly
        sage: growth = diagonal_asy(1/(1 - x - y), output_format="asymptotic"); growth
        1/sqrt(pi)*4^n*n^(-1/2) + O(4^n*n^(-3/2))
        sage: growth.parent()
        Asymptotic Ring <(Algebraic Real Field)^n * n^QQ * (Arg_(Algebraic Field))^n> over Symbolic Ring

    Increasing the precision of the expansion returns an expansion with more terms
    (works for all available output formats)::

        sage: diagonal_asy(1/(1 - x - y), expansion_precision=3, output_format="asymptotic")
        1/sqrt(pi)*4^n*n^(-1/2) - 1/8/sqrt(pi)*4^n*n^(-3/2) + 1/128/sqrt(pi)*4^n*n^(-5/2)
        + O(4^n*n^(-7/2))

    The direction of the diagonal, `r`, defaults to the standard diagonal (i.e., the
    vector of all 1's) if not specified. It also supports passing non-integer values,
    notably rational numbers::

        sage: diagonal_asy(1/(1 - x - y), r=(1, 17/42), output_format="symbolic")
        1.317305628032865?*2.324541507270374?^n/(sqrt(pi)*sqrt(n))

    and even algebraic numbers (note, however, that the performance for complicated
    algebraic numbers is significantly degraded)::

        sage: diagonal_asy(1/(1 - x - y), r=(sqrt(2), 1))
        0.9238795325112868?/sqrt(pi)*(2.414213562373095?/0.5857864376269049?^1.414213562373095?)^n*n^(-1/2) + O((2.414213562373095?/0.5857864376269049?^1.414213562373095?)^n*n^(-3/2))

    ::

        sage: diagonal_asy(1/(1 - x - y*x^2), r=(1, 1/2 - 1/2*sqrt(1/5)), output_format="asymptotic")
        1.710862642974252?/sqrt(pi)*1.618033988749895?^n*n^(-1/2)
        + O(1.618033988749895?^n*n^(-3/2))

    The function times individual steps of the algorithm, timings can
    be displayed by increasing the printed verbosity level of our debug logger::

        sage: import logging
        sage: from sage_acsv import ACSVSettings
        sage: ACSVSettings.set_logging_level(logging.INFO)
        sage: diagonal_asy(1/(1 - x - y))
        INFO:sage_acsv:... Executed Kronecker in ... seconds.
        INFO:sage_acsv:... Executed Minimal Points in ... seconds.
        INFO:sage_acsv:... Executed Final Asymptotics in ... seconds.
        1/sqrt(pi)*4^n*n^(-1/2) + O(4^n*n^(-3/2))
        sage: ACSVSettings.set_logging_level(logging.WARNING)

    Extraction of coefficient asymptotics even works in cases where the singular variety of `F`
    is not smooth::

        sage: diagonal_asy(1/((1-(2*x+y)/3)*(1-(3*x+y)/4)), r = [17/24, 7/24], output_format = 'asymptotic')
        12 + O(n^(-1))

        sage: diagonal_asy(1/((1-(2*x+y)/3)*(1-(3*x+y)/4)), r = [17/24, 7/24], output_format = 'asymptotic')
        12 + O(n^(-1))
        sage: G = (1+x)*(1-x*y^2+x^2)
        sage: H = (1-z*(1+x^2+x*y^2))*(1-y)*(1+x^2)
        sage: strat = [
        ....:     [(1-z*(1+x^2+x*y^2), 1-y, 1+x^2)],
        ....:     [(1-z*(1+x^2+x*y^2), 1-y),(1-z*(1+x^2+x*y^2), 1+x^2),(1-y,1+x^2)],
        ....:     [(H,)],
        ....: ]
        sage: diagonal_asy(G/H, r = [1,1,1], output_format = 'asymptotic', whitney_strat = strat)
        0.866025403784439?/sqrt(pi)*3^n*n^(-1/2) + O(3^n*n^(-3/2))

    TESTS:

    Check that the workaround for the AsymptoticRing swallowing
    the modulus works as intended::

        sage: diagonal_asy(1/(1 - x^4 - y^4))  # long time
        1/2/sqrt(pi)*1.414213562373095?^n*n^(-1/2) + 1/2/sqrt(pi)*1.414213562373095?^n*n^(-1/2)*(e^(I*arg(-1)))^n + 1/2/sqrt(pi)*1.414213562373095?^n*n^(-1/2)*(e^(I*arg(-I)))^n + 1/2/sqrt(pi)*1.414213562373095?^n*n^(-1/2)*(e^(I*arg(I)))^n + O(1.414213562373095?^n*n^(-3/2))

    Check that there are no prohibited variable names::

        sage: var('n t u_')
        (n, t, u_)
        sage: diagonal_asy(1/(1 - n - t - u_))
        0.866025403784439?/pi*27^n*n^(-1) + O(27^n*n^(-2))

    """
    G, H, variable_map = _prepare_symbolic_fraction(F)
    vs = list(variable_map.values())

    if whitney_strat is not None:
        whitney_strat = [
            [
                tuple(SR(gen).subs(variable_map) for gen in component)
                for component in stratum
            ]
            for stratum in whitney_strat
        ]
    if linear_form is not None:
        linear_form = SR(linear_form).subs(variable_map)

    if r is None:
        n = len(H.variables())
        r = [1 for _ in range(n)]

    try:
        r = [QQ(ri) for ri in r]
    except (ValueError, TypeError):
        r = [AA(ri) for ri in r]

    R = PolynomialRing(QQ, vs, len(vs))
    H_sing = Ideal([R(H)] + [R(H.derivative(v)) for v in vs])
    if H_sing.dimension() < 0:
        return _diagonal_asy_smooth(
            G,
            H,
            r=r,
            linear_form=linear_form,
            expansion_precision=expansion_precision,
            return_points=return_points,
            output_format=output_format,
            as_symbolic=as_symbolic,
        )

    t, lambda_, u_ = PolynomialRing(QQ, "t, lambda_, u_").gens()
    expanded_R = PolynomialRing(QQ, len(vs) + 3, vs + [t, lambda_, u_])

    vs = [expanded_R(v) for v in vs]
    t, lambda_, u_ = expanded_R(t), expanded_R(lambda_), expanded_R(u_)

    d = len(vs)

    # Make sure G and H are coprime, and that H does not vanish at 0
    G, H = RationalFunctionReduce(G, H)
    G, H = expanded_R(G), expanded_R(H)
    if H.subs({v: 0 for v in H.variables()}) == 0:
        raise ValueError("Denominator vanishes at 0.")

    H_sf = prod([f for f, _ in H.factor()])
    # In case form doesn't separate, we want to try again
    for _ in range(ACSVSettings.MAX_MIN_CRIT_RETRIES):
        try:
            # Find minimal critical points in Kronecker Representation
            min_crit_pts = _find_contributing_points_combinatorial(
                G, H_sf, vs, r=r, linear_form=linear_form, whitney_strat=whitney_strat
            )
            break
        except Exception as e:
            if isinstance(e, ACSVException) and e.retry:
                acsv_logger.warning(
                    "Randomly generated linear form was not suitable, "
                    f"encountered error: {e}\nRetrying..."
                )
                continue
            else:
                raise e
    else:
        return

    asm_quantities = []
    for cp in min_crit_pts:
        # Step 1: Determine if pt is a transverse multiple point of H, and compute the factorization
        # for now, we'll just try to factor it in the polynomial ring
        R = PolynomialRing(QQbar, len(vs), vs)
        G = R(SR(G))
        H = R(SR(H))
        vs = [R(SR(v)) for v in vs]
        subs_dict = {vs[i]: cp[i] for i in range(d)}
        poly_factors = H.factor()
        unit = poly_factors.unit()
        factors = []
        multiplicities = []
        for factor, multiplicity in poly_factors:
            const = factor.coefficients()[-1]
            unit *= const**multiplicity
            factor /= const
            if factor.subs(subs_dict) != 0:
                unit *= factor.subs(subs_dict)
                continue
            factors.append(factor)
            multiplicities.append(multiplicity)
        s = len(factors)
        normals = matrix(
            [[f.derivative(v).subs(subs_dict) for v in vs] for f in factors]
        )
        if normals.rank() < s:
            raise ACSVException(
                "Not a transverse intersection. Cannot deal with this case."
            )

        # Step 2: Find the locally parametrizing coordinates of the point pt
        # Since we have d variables and s factors, there should be d-s of these parametrizing coordinates
        # We will try to parametrize with the first d-s coordinates, shuffling the vs and r if it doesn't work
        for _ in range(s**2):
            Jac = matrix(
                [
                    [(v * Q.derivative(v)).subs(subs_dict) for v in vs[d - s :]]
                    for Q in factors
                ]
            )
            if Jac.determinant() != 0:
                break

            acsv_logger.info("Variables do not parametrize, shuffling")
            vs_r_cp = list(zip(vs, r, cp))
            shuffle(vs_r_cp)  # shuffle mutates the list
            vs, r, cp = zip(*vs_r_cp)
        else:
            raise ACSVException("Cannot find parametrizing set.")

        # Step 3: Compute the gamma matrix as defined in 9.10
        Gamma = matrix(
            [[(v * Q.derivative(v)).subs(subs_dict) for v in vs] for Q in factors]
            + [
                [v.subs(subs_dict) if vs.index(v) == i else 0 for i in range(d)]
                for v in vs[: d - s]
            ]
        )

        # Some constants appearing for higher order singularities
        mult_fac = prod([factorial(m - 1) for m in multiplicities])
        r_gamma_inv = prod(
            x ** (multiplicities[i] - 1)
            for i, x in enumerate(list(vector(r) * Gamma.inverse())[:s])
        )
        # If cp lies on a single smooth component, we can compute asymptotics like in the smooth case
        if s == 1 and sum(multiplicities) == 1:
            n = SR.var("n")
            expansion = sum(
                term / (r[-1] * n) ** (term_order)
                for term_order, term in enumerate(
                    GeneralTermAsymptotics(G, H, r, vs, cp, expansion_precision)
                )
            )
            Det = GetHessian(H, vs, r).determinant()
            B = SR(1 / Det.subs(subs_dict) / r[-1] ** (d - 1) / 2 ** (d - 1))
        else:
            # Higher order expansions not currently supported for non-smooth critical points
            if expansion_precision > 1:
                acsv_logger.warning(
                    "Higher order expansions are not supported in the non-smooth case. Defaulting to expansion_precision 1."
                )
            # For non-complete intersections, we must compute the parametrized Hessian matrix
            if s != d:
                Qw = ImplicitHessian(factors, vs, r, subs=subs_dict)
                expansion = SR(G.subs(subs_dict) / abs(Gamma.determinant()) / unit)
                B = SR(
                    prod([v for v in vs[: d - s]]).subs(subs_dict)
                    / (r[-1] * Qw).determinant()
                    / 2 ** (d - s)
                )
            else:
                expansion = SR(G.subs(subs_dict) / unit / abs(Gamma.determinant()))
                B = 1

        expansion *= (
            (-1) ** sum([m - 1 for m in multiplicities]) * r_gamma_inv / mult_fac
        )

        T = prod(SR(vs[i].subs(subs_dict)) ** r[i] for i in range(d))
        C = SR(1 / T)
        D = QQ((s - d) / 2 + sum(multiplicities) - s)
        try:
            B = QQbar(B)
            C = QQbar(C)
        except (ValueError, TypeError):
            pass

        asm_quantities.append([expansion, B, C, D, s])

    asm_vals = [(c, d, b.sqrt(), a, s) for a, b, c, d, s in asm_quantities]

    if as_symbolic:
        acsv_logger.warning(
            "The as_symbolic argument has been deprecated in favor of output_format='symbolic' "
        )
        output_format = ACSVSettings.Output.SYMBOLIC

    if output_format is None:
        output_format = ACSVSettings.get_default_output_format()
    else:
        output_format = ACSVSettings.Output(output_format)

    if output_format in (ACSVSettings.Output.TUPLE, ACSVSettings.Output.SYMBOLIC):
        n = SR.var("n")
        result = [
            (base, n**exponent, (pi ** (s - d)).sqrt(), constant * expansion)
            for (base, exponent, constant, expansion, s) in asm_vals
        ]
        if output_format == ACSVSettings.Output.SYMBOLIC:
            result = sum([a**n * b * c * d for (a, b, c, d) in result])

    elif output_format == ACSVSettings.Output.ASYMPTOTIC:
        from sage.all import AsymptoticRing

        AR = AsymptoticRing("QQbar^n * n^QQ", QQbar)
        n = AR.gen()
        result = sum(
            [  # bug in AsymptoticRing requires splitting out modulus manually
                constant
                * (pi ** (s - d)).sqrt()
                * abs(base) ** n
                * collapse_zero_part(base / abs(base)) ** n
                * n**exponent
                * AR(expansion)
                + (abs(base) ** n * n ** (exponent - expansion_precision)).O()
                for (base, exponent, constant, expansion, s) in asm_vals
            ]
        )

    else:
        raise NotImplementedError(f"Missing implementation for {output_format}")

    if return_points:
        return result, min_crit_pts

    return result


def GeneralTermAsymptotics(G, H, r, vs, cp, expansion_precision):
    r"""
    Compute coefficients of general (not necessarily leading) terms of
    the asymptotic expansion for a given critical
    point of a rational combinatorial multivariate rational function.

    Typically, this function is called as a subroutine of :func:`.diagonal_asy`.

    INPUT:

    * ``G, H`` -- Coprime polynomials with `F = G/H`.
    * ``vs`` -- Tuple of variables occurring in `G` and `H`.
    * ``r`` -- The direction. A length `d` vector of positive algebraic numbers (usually
      integers).
    * ``cp`` -- A minimal critical point of `F` with coordinates specified in the
      same order as in ``vs``.
    * ``expansion_precision`` -- A positive integer value. This is the number of terms
      for which to compute coefficients in the asymptotic expansion.

    OUTPUT:

    List of coefficients of the asymptotic expansion.

    EXAMPLES::

        sage: from sage_acsv import GeneralTermAsymptotics
        sage: R.<x, y, z> = QQ[]
        sage: GeneralTermAsymptotics(1, 1 - x - y, [1, 1], [x, y], [1/2, 1/2], 5)
        [2, -1/4, 1/64, 5/512, -21/16384]
        sage: GeneralTermAsymptotics(1, 1 - x - y - z, [1, 1, 1], [x, y, z], [1/3, 1/3, 1/3], 4)
        [3, -2/3, 2/27, 14/729]
    """

    if expansion_precision == 1:
        A = SR(-G / vs[-1] / H.derivative(vs[-1]))
        subs_dict = {SR(v): V for (v, V) in zip(vs, cp)}
        return [A.subs(subs_dict)]

    # Convert everything to field of algebraic numbers
    d = len(vs)
    R = PolynomialRing(QQbar, vs)
    vs = R.gens()
    vd = vs[-1]
    tvars = SR.var("t", d - 1)
    G, H = R(SR(G)), R(SR(H))

    cp = {v: V for (v, V) in zip(vs, cp)}

    W = DifferentialWeylAlgebra(PolynomialRing(QQbar, tvars))
    TR = QQbar[[tvars]]
    T = TR.gens()
    tvars = T
    D = list(W.differentials())

    # Function to apply differential operator dop on function f
    def eval_op(dop, f):
        if len(f.parent().gens()) == 1:
            return sum(
                prod([factorial(k) for k in E[0][1]]) * E[1] * f[E[0][1][0]]
                for E in dop
            )
        else:
            return sum(
                [prod([factorial(k) for k in E[0][1]]) * E[1] * f[E[0][1]] for E in dop]
            )

    Hess = GetHessian(H, vs, r, cp)
    Hessinv = Hess.inverse()
    v = matrix(W, [D[: d - 1]])
    Epsilon = -(v * Hessinv.change_ring(W) * v.transpose())[0, 0]

    # P and PsiTilde only need to be computed to order 2M
    N = 2 * expansion_precision + 1

    # Find series expansion of function g given implicitly by
    # H(w_1, ..., w_{d-1}, g(w_1, ..., w_{d-1})) = 0 up to needed order
    g = NewtonSeries(H.subs({v: v + cp[v] for v in vs}), vs, N)
    g = g.subs({v: v - cp[v] for v in vs}) + cp[vd]

    # Polar change of coordinates
    tsubs = {v: cp[v] * exp(I * t).add_bigoh(N) for v, t in zip(vs, tvars)}
    tsubs[vd] = g.subs(tsubs)

    # Compute PsiTilde up to needed order
    psi = log(g.subs(tsubs) / g.subs(cp)).add_bigoh(N)
    psi += I * sum([r[k] * tvars[k] for k in range(d - 1)]) / r[-1]
    v = matrix(TR, [tvars[k] for k in range(d - 1)])
    psiTilde = psi - (v * Hess * v.transpose())[0, 0] / 2
    PsiSeries = psiTilde.truncate(N)

    # Compute series expansion of P = -G/(g*H_{z_d}) up to needed order
    P_num = -G.subs(tsubs).add_bigoh(N)
    P_denom = (g * H.derivative(vd)).subs(tsubs).add_bigoh(N)
    PSeries = (P_num / P_denom).truncate(N)

    if len(tvars) > 1:
        PsiSeries = PsiSeries.polynomial()
        PSeries = PSeries.polynomial()

    # Precompute products used for asymptotics
    EE = [Epsilon**k for k in range(3 * expansion_precision - 2)]
    PP = [PSeries]
    for k in range(1, 2 * expansion_precision - 1):
        PP.append(PP[k - 1] * PsiSeries)

    # Function to compute constants appearing in asymptotic expansion
    def constants_clj(ell, j):
        extra_contrib = (-1) ** j / (
            2 ** (ell + j) * factorial(ell) * factorial(ell + j)
        )
        return extra_contrib * eval_op(EE[ell + j], PP[ell])

    res = [
        sum([constants_clj(ell, j) for ell in srange(2 * j + 1)])
        for j in srange(expansion_precision)
    ]
    try:
        for i in range(len(res)):
            if res[i].imag() == 0:
                res[i] = AA(res[i])
    except (TypeError, ValueError, NotImplementedError):
        pass

    return res


def ContributingCombinatorialSmooth(G, H, variables, r=None, linear_form=None):
    r"""Compute contributing points of a combinatorial multivariate
    rational function `F=G/H` admitting a finite number of critical points.
    Assumes the singular variety of `F` is smooth.

    Typically, this function is called as a subroutine of :func:`._diagonal_asy_smooth`.

    INPUT:

    * ``G, H`` -- Coprime polynomials with `F = G/H`.
    * ``variables`` -- List of variables of ``G`` and ``H``.
    * ``r`` -- (Optional) the direction, a vector of positive algebraic
      numbers (usually integers).
    * ``linear_form`` -- (Optional) A linear combination of the input
      variables that separates the critical point solutions.

    OUTPUT:

    List of minimal critical points of `F` in the direction `r`,
    as a list of tuples of algebraic numbers.

    NOTE:

    The code randomly generates a linear form, which for generic rational functions
    separates the solutions of an intermediate polynomial system with high probability.
    This separation step can fail, but (assuming F has a finite number of critical points)
    the code can be rerun until a separating form is found.

    Examples::

        sage: from sage_acsv import ContributingCombinatorialSmooth
        sage: R.<x, y, w, lambda_, t, u_> = QQ[]
        sage: pts = ContributingCombinatorialSmooth(
        ....:     1,
        ....:     1 - w*(y + x + x^2*y + x*y^2),
        ....:     [w, x, y],
        ....: )
        sage: sorted(pts)
        [[-1/4, -1, -1], [1/4, 1, 1]]
    """

    timer = Timer()
    (
        expanded_R,
        vs,
        (t, lambda_, u_),
        r,
        r_variable_values,
    ) = _prepare_expanded_polynomial_ring(variables, direction=r)

    G, H = expanded_R(G), expanded_R(H)
    vsT = vs + list(r_variable_values.keys()) + [t, lambda_]

    # Create the critical point equations system
    vsH = H.variables()
    system = [
        H_var * H.derivative(H_var) - r_var * lambda_
        for H_var, r_var in zip(H.variables(), r)
    ]
    system.extend([H, H.subs({z: z * t for z in vsH})])
    system.extend(
        [
            direction_value.minpoly().subs(direction_var)
            for direction_var, direction_value in r_variable_values.items()
        ]
    )

    # Compute the Kronecker representation of our system
    timer.checkpoint()

    P, Qs = _kronecker_representation(system, u_, vsT, linear_form)
    timer.checkpoint("Kronecker")

    Qt = Qs[-2]  # Qs ordering is H.variables() + rvars + [t, lambda_]
    Pd = P.derivative()

    # Solutions to Pt are solutions to the system where t is not 1
    one_minus_t = gcd(Pd - Qt, P)
    Pt, _ = P.quo_rem(one_minus_t)
    rts_t_zo = list(
        filter(
            lambda k: (Qt / Pd).subs(u_=k) > 0 and (Qt / Pd).subs(u_=k) < 1,
            Pt.roots(AA, multiplicities=False),
        )
    )
    non_min = [[(q / Pd).subs(u_=u) for q in Qs[0:-2]] for u in rts_t_zo]

    # Filter the real roots for minimal points with positive coords
    pos_minimals = []
    for u in one_minus_t.roots(AA, multiplicities=False):
        is_min = True
        v = [(q / Pd).subs(u_=u) for q in Qs[: len(vs)]]
        rv = {
            ri: (q / Pd).subs(u_=u)
            for (ri, q) in zip(r_variable_values, Qs[len(vs) : -2])
        }
        if any([rv[ri] != ri_value for ri, ri_value in r_variable_values.items()]):
            continue
        if any([value <= 0 for value in v[: len(vs)]]):
            continue
        for pt in non_min:
            if all([a == b for (a, b) in zip(v, pt)]):
                is_min = False
                break
        if is_min:
            pos_minimals.append(u)

    # Remove non-smooth points and points with zero coordinates (where lambda=0)
    for i in range(len(pos_minimals)):
        x = (Qs[-1] / Pd).subs(u_=pos_minimals[i])
        if x == 0:
            acsv_logger.warning(
                f"Removing critical point {pos_minimals[i]} because it either "
                "has a zero coordinate or is not smooth."
            )
            pos_minimals.pop(i)

    # Verify necessary assumptions
    if len(pos_minimals) == 0:
        raise ACSVException("No smooth minimal critical points found.")
    elif len(pos_minimals) > 1:
        raise ACSVException(
            f"More than one minimal point with positive real coordinates found: {pos_minimals}"
        )

    # Find all minimal critical points
    minCP = [(q / Pd).subs(u_=pos_minimals[0]) for q in Qs[0:-2]]
    minimals = []

    for u in one_minus_t.roots(QQbar, multiplicities=False):
        v = [(q / Pd).subs(u_=u) for q in Qs[: len(vs)]]
        rv = {
            ri: (q / Pd).subs(u_=u)
            for (ri, q) in zip(r_variable_values, Qs[len(vs) : -2])
        }
        if any([rv[r_var] != r_value for r_var, r_value in r_variable_values.items()]):
            continue
        if all([a.abs() == b.abs() for (a, b) in zip(minCP, v)]):
            minimals.append(u)

    # Get minimal point coords, and make exact if possible
    minimal_coords = [[(q / Pd).subs(u_=u) for q in Qs[: len(vs)]] for u in minimals]
    [[a.exactify() for a in b] for b in minimal_coords]

    timer.checkpoint("Minimal Points")

    return [[(q / Pd).subs(u_=u) for q in Qs[: len(vs)]] for u in minimals]


def _find_contributing_points_combinatorial(
    G,
    H,
    variables,
    r=None,
    linear_form=None,
    whitney_strat=None,
):
    r"""Compute contributing points of a combinatorial multivariate
    rational function `F=G/H` admitting a finite number of critical points.

    Typically, this function is called as a subroutine of :func:`.diagonal_asy`.

    INPUT:

    * ``G, H`` -- Coprime polynomials with ``F = G/H``
    * ``variables`` -- List of variables of ``G`` and ``H``
    * ``r`` -- (Optional) Length ``d`` vector of positive integers
    * ``linear_form`` -- (Optional) A linear combination of the input
      variables that separates the critical point solutions
    * ``whitney_strat`` -- (Optional) If known / precomputed, a
      Whitney Stratification of `V(H)`. The program will not check if
      this stratification is correct. Should be a list of length ``d``, where
      the ``k``-th entry is a list of tuples of ideas generators representing
      a component of the ``k``-dimensional stratum.

    OUTPUT:

    A list of minimal contributing points of `F` in the direction `r`.

    NOTE:

    The code randomly generates a linear form, which for generic rational functions
    separates the solutions of an intermediate polynomial system with high probability.
    This separation step can fail, but (assuming F has a finite number of critical points)
    the code can be rerun until a separating form is found.
    """
    (
        expanded_R,
        vs,
        (t, lambda_, u_),
        r,
        r_variable_values,
    ) = _prepare_expanded_polynomial_ring(variables, direction=r)
    H = expanded_R(H)
    vsT = vs + list(r_variable_values.keys()) + [t, lambda_]

    # Compute the critical point system for each stratum
    pure_H = PolynomialRing(QQ, vs)

    if whitney_strat is None:
        whitney_strat = WhitneyStrat(Ideal(pure_H(H)), pure_H)
    else:
        # Cast symbolic generators for provided stratification into the correct ring
        whitney_strat = [
            prod([Ideal([pure_H(f) for f in comp]) for comp in stratum])
            for stratum in whitney_strat
        ]

    critical_point_ideals = []
    for d, stratum in enumerate(whitney_strat):
        critical_point_ideals.append([])
        for P in PrimaryDecomposition(stratum):
            c = len(vs) - d
            P_ext = P.change_ring(expanded_R)
            M = matrix([[v * f.derivative(v) for v in vs] for f in P_ext.gens()] + [r])
            # Add in min polys for the direction variables
            r_polys = [
                direction_value.minpoly().subs(direction_var)
                for direction_var, direction_value in r_variable_values.items()
            ]
            # Create ideal of expanded_R containing extended critical point equations
            cpid = P_ext + Ideal(
                M.minors(c + 1)
                + [H.subs({v: v * t for v in vs}), (prod(vs) * lambda_ - 1)]
                + r_polys
            )
            # Saturate cpid by lower dimension stratum, if d > 0
            if d > 0:
                cpid = Saturate(cpid, whitney_strat[d - 1].change_ring(expanded_R))

            critical_point_ideals[-1].append((P, cpid))

    # Final minimal critical points with positive coordinates on each stratum
    critical_points_by_stratum = {}
    pos_minimals_by_stratum = {}
    for d in reversed(range(len(critical_point_ideals))):
        ideals = critical_point_ideals[d]
        critical_points_by_stratum[d] = []
        pos_minimals_by_stratum[d] = []

        for _, ideal in ideals:
            if ideal.dimension() < 0:
                continue
            P, Qs = _kronecker_representation(ideal.gens(), u_, vsT, linear_form)

            Qt = Qs[-2]  # Qs ordering is H.variables() + rvars + [t, lambda_]
            Pd = P.derivative()

            # Solutions to Pt are solutions to the system where t is not 1
            one_minus_t = gcd(Pd - Qt, P)
            Pt, _ = P.quo_rem(one_minus_t)
            rts_t_zo = list(
                filter(
                    lambda k: (Qt / Pd).subs(u_=k) > 0 and (Qt / Pd).subs(u_=k) < 1,
                    Pt.roots(AA, multiplicities=False),
                )
            )
            non_min = [[(q / Pd).subs(u_=u) for q in Qs[0:-2]] for u in rts_t_zo]

            # Filter the real roots for minimal points with positive coords
            pos_minimals = []
            for u in one_minus_t.roots(AA, multiplicities=False):
                is_min = True
                v = [(q / Pd).subs(u_=u) for q in Qs[: len(vs)]]
                rv = {
                    ri: (q / Pd).subs(u_=u)
                    for (ri, q) in zip(r_variable_values, Qs[len(vs) : -2])
                }
                if any(
                    [rv[ri] != ri_value for ri, ri_value in r_variable_values.items()]
                ):
                    continue
                if any([value <= 0 for value in v[: len(vs)]]):
                    continue
                for pt in non_min:
                    if all([a == b for (a, b) in zip(v, pt)]):
                        is_min = False
                        break
                if is_min:
                    pos_minimals.append(u)

            pos_minimals_by_stratum[d].extend(
                [
                    [
                        collapse_zero_part(QQbar((q / Pd).subs(u_=u)))
                        for q in Qs[: len(vs)]
                    ]
                    for u in pos_minimals
                ]
            )

            # Characterize all complex critical points in each stratum
            for u in one_minus_t.roots(QQbar, multiplicities=False):
                rv = {
                    ri: (q / Pd).subs(u_=u)
                    for (ri, q) in zip(r_variable_values, Qs[len(vs) : -2])
                }
                if any(
                    rv[r_var] != r_value for r_var, r_value in r_variable_values.items()
                ):
                    continue

                w = [
                    collapse_zero_part(QQbar((q / Pd).subs(u_=u)))
                    for q in Qs[: len(vs)]
                ]
                critical_points_by_stratum[d].append(w)

    # Refine positive minimal critical points to those that are contributing
    contributing_pos_minimals = []
    all_factors = list(factor[0] for factor in H.factor())
    r = [QQbar(r_variable_values.get(ri, ri)) for ri in r]
    for d in reversed(range(len(critical_point_ideals))):
        pos_minimals = pos_minimals_by_stratum[d]
        if len(contributing_pos_minimals) > 0:
            break

        for x in pos_minimals:
            if IsContributing(vs, x, r, all_factors, len(vs) - d):
                contributing_pos_minimals.append(x)
                for i in range(d):
                    stratum = whitney_strat[i]
                    if stratum.subs(
                        {pure_H(wi): val for wi, val in zip(vs, x)}
                    ) == Ideal(pure_H(0)):
                        raise ACSVException(
                            "Non-generic direction detected - critical point {w} is contained in {dim}-dimensional stratum".format(
                                w=str(x), dim=i
                            )
                        )

    if len(contributing_pos_minimals) == 0:
        raise ACSVException("No contributing points found.")
    if len(contributing_pos_minimals) > 1:
        raise ACSVException(
            f"More than one minimal contributing point with positive real coordinates found: {contributing_pos_minimals}"
        )
    minimal = contributing_pos_minimals[0]

    # Characterize all complex contributing points
    contributing_points = []
    for d in reversed(range(len(critical_point_ideals))):
        for w in critical_points_by_stratum[d]:
            if all(
                [abs(w_i) == abs(min_i) for w_i, min_i in zip(w, minimal)]
            ) and IsContributing(vs, w, r, all_factors, len(vs) - d):
                contributing_points.append(w)

    return contributing_points


def ContributingCombinatorial(
    F,
    r=None,
    linear_form=None,
    whitney_strat=None,
):
    r"""Compute contributing points of a combinatorial multivariate
    rational function `F=G/H` admitting a finite number of critical points.

    INPUT:

    * ``F`` -- Symbolic fraction, the rational function assumed to have
      a finite number of critical points
    * ``r`` -- (Optional) Length ``d`` vector of positive integers
    * ``linear_form`` -- (Optional) A linear combination of the input
      variables that separates the critical point solutions
    * ``whitney_strat`` -- (Optional) If known / precomputed, a
      Whitney Stratification of `V(H)`. The program will not check if
      this stratification is correct. Should be a list of length ``d``, where
      the ``k``-th entry is a list of tuples of ideas generators representing
      a component of the ``k``-dimensional stratum.

    OUTPUT:

    A list of minimal contributing points of `F` in the direction `r`.

    NOTE:

    The code randomly generates a linear form, which for generic rational functions
    separates the solutions of an intermediate polynomial system with high probability.
    This separation step can fail, but (assuming F has a finite number of critical points)
    the code can be rerun until a separating form is found.

    Examples::

        sage: from sage_acsv import ContributingCombinatorial
        sage: var('x y')
        (x, y)
        sage: pts = ContributingCombinatorial(1/((1-(2*x+y)/3)*(1-(3*x+y)/4)))
        sage: sorted(pts)
        [[3/4, 3/2]]

    """
    G, H, variable_map = _prepare_symbolic_fraction(F)
    variables = list(variable_map.values())
    if whitney_strat is not None:
        whitney_strat = [
            [
                tuple(SR(gen).subs(variable_map) for gen in component)
                for component in stratum
            ]
            for stratum in whitney_strat
        ]
    if linear_form is not None:
        linear_form = SR(linear_form).subs(variable_map)

    return _find_contributing_points_combinatorial(
        G,
        H,
        variables,
        r=r,
        linear_form=linear_form,
        whitney_strat=whitney_strat,
    )


def MinimalCriticalCombinatorial(F, r=None, linear_form=None, whitney_strat=None):
    r"""Compute nonzero minimal critical points of a combinatorial multivariate
    rational function `F=G/H` admitting a finite number of critical points.

    INPUT:

    * ``F`` -- Symbolic fraction, the rational function of interest.
    * ``r`` -- (Optional) Length ``d`` vector of positive integers
    * ``linear_form`` -- (Optional) A linear combination of the input
      variables that separates the critical point solutions
    * ``whitney_strat`` -- (Optional) If known / precomputed, a
      Whitney Stratification of `V(H)`. The program will not check if
      this stratification is correct. Should be a list of length ``d``, where
      the ``k``-th entry is a list of tuples of ideas generators representing
      a component of the ``k``-dimensional stratum.

    OUTPUT:

    A list of minimal contributing points of `F` in the direction `r`.

    NOTE:

    The code randomly generates a linear form, which for generic rational functions
    separates the solutions of an intermediate polynomial system with high probability.
    This separation step can fail, but (assuming F has a finite number of critical points)
    the code can be rerun until a separating form is found.

    Examples::

        sage: from sage_acsv import MinimalCriticalCombinatorial
        sage: var('x y')
        (x, y)
        sage: pts = MinimalCriticalCombinatorial(1/((1-(2*x+y)/3)*(1-(3*x+y)/4)))
        sage: sorted(pts)
        [[3/4, 3/2], [1, 1]]

    """
    _, H, variable_map = _prepare_symbolic_fraction(F)
    variables = list(variable_map.values())
    if whitney_strat is not None:
        whitney_strat = [
            [
                tuple(SR(gen).subs(variable_map) for gen in component)
                for component in stratum
            ]
            for stratum in whitney_strat
        ]
    if linear_form is not None:
        linear_form = SR(linear_form).subs(variable_map)
    (
        expanded_R,
        vs,
        (t, lambda_, u_),
        r,
        r_variable_values,
    ) = _prepare_expanded_polynomial_ring(variables, direction=r)
    H = expanded_R(H)

    vsT = vs + list(r_variable_values) + [t, lambda_]

    # Compute the critical point system for each stratum
    pure_H = PolynomialRing(QQ, vs)

    if whitney_strat is None:
        whitney_strat = WhitneyStrat(Ideal(pure_H(H)), pure_H)
    else:
        # Cast symbolic generators for provided stratification into the correct ring
        whitney_strat = [
            prod([Ideal([pure_H(f) for f in comp]) for comp in stratum])
            for stratum in whitney_strat
        ]

    critical_points = []
    pos_minimals = []
    for d, stratum in enumerate(whitney_strat):
        for P_comp in PrimaryDecomposition(stratum):
            c = len(vs) - d
            P_ext = P_comp.change_ring(expanded_R)
            M = matrix([[v * f.derivative(v) for v in vs] for f in P_ext.gens()] + [r])
            # Add in min polys for the direction variables
            r_polys = [
                direction_value.minpoly().subs(direction_var)
                for direction_var, direction_value in r_variable_values.items()
            ]
            # Create ideal of expanded_R containing extended critical point equations
            ideal = P_ext + Ideal(
                M.minors(c + 1)
                + [H.subs({v: v * t for v in vs}), (prod(vs) * lambda_ - 1)]
                + r_polys
            )
            # Saturate cpid by lower dimension stratum, if d > 0
            if d > 0:
                ideal = Saturate(ideal, whitney_strat[d - 1].change_ring(expanded_R))

            if ideal.dimension() < 0:
                continue
            P, Qs = _kronecker_representation(ideal.gens(), u_, vsT, linear_form)

            Qt = Qs[-2]  # Qs ordering is H.variables() + rvars + [t, lambda_]
            Pd = P.derivative()

            # Solutions to Pt are solutions to the system where t is not 1
            one_minus_t = gcd(Pd - Qt, P)
            Pt, _ = P.quo_rem(one_minus_t)
            rts_t_zo = list(
                filter(
                    lambda k: (Qt / Pd).subs(u_=k) > 0 and (Qt / Pd).subs(u_=k) < 1,
                    Pt.roots(AA, multiplicities=False),
                )
            )
            non_min = [[(q / Pd).subs(u_=u) for q in Qs[0:-2]] for u in rts_t_zo]

            # Filter the real roots for minimal points with positive coords
            for u in one_minus_t.roots(AA, multiplicities=False):
                is_min = True
                v = [(q / Pd).subs(u_=u) for q in Qs[: len(vs)]]
                rv = {
                    ri: (q / Pd).subs(u_=u)
                    for (ri, q) in zip(r_variable_values, Qs[len(vs) : -2])
                }
                if any(
                    [rv[ri] != ri_value for ri, ri_value in r_variable_values.items()]
                ):
                    continue
                if any([value <= 0 for value in v[: len(vs)]]):
                    continue
                for pt in non_min:
                    if all([a == b for (a, b) in zip(v, pt)]):
                        is_min = False
                        break
                if is_min:
                    pos_minimals.append(v)

            # Characterize all complex critical points in each stratum
            for u in one_minus_t.roots(QQbar, multiplicities=False):
                rv = {
                    ri: (q / Pd).subs(u_=u)
                    for (ri, q) in zip(r_variable_values, Qs[len(vs) : -2])
                }
                if any(
                    rv[r_var] != r_value for r_var, r_value in r_variable_values.items()
                ):
                    continue

                w = [QQbar((q / Pd).subs(u_=u)) for q in Qs[: len(vs)]]
                critical_points.append(w)

    if len(pos_minimals) == 0:
        raise ACSVException("No critical points found.")

    # Characterize all complex contributing points
    minimal_criticals = []
    for w in critical_points:
        if any(
            all([abs(w_i) == abs(min_i) for w_i, min_i in zip(w, minimal)])
            for minimal in pos_minimals
        ):
            minimal_criticals.append(w)

    minimal_criticals = [
        [collapse_zero_part(w_i) for w_i in w] for w in minimal_criticals
    ]
    return minimal_criticals


def CriticalPoints(F, r=None, linear_form=None, whitney_strat=None):
    r"""Compute critical points of a multivariate
    rational function `F=G/H` admitting a finite number of critical points.

    Typically, this function is called as a subroutine of :func:`.diagonal_asy`.

    INPUT:

    * ``F`` -- Symbolic fraction, the rational function of interest.
    * ``r`` -- (Optional) Length ``d`` vector of positive integers
    * ``linear_form`` -- (Optional) A linear combination of the input
      variables that separates the critical point solutions
    * ``whitney_strat`` -- (Optional) If known / precomputed, a
      Whitney Stratification of `V(H)`. The program will not check if
      this stratification is correct. Should be a list of length ``d``, where
      the ``k``-th entry is a list of tuples of ideas generators representing
      a component of the ``k``-dimensional stratum.

    OUTPUT:

    A list of minimal contributing points of `F` in the direction `r`,

    NOTE:

    The code randomly generates a linear form, which for generic rational functions
    separates the solutions of an intermediate polynomial system with high probability.
    This separation step can fail, but (assuming F has a finite number of critical points)
    the code can be rerun until a separating form is found.

    Examples::

        sage: from sage_acsv import CriticalPoints
        sage: var('x y')
        (x, y)
        sage: pts = CriticalPoints(1/((1-(2*x+y)/3)*(1-(3*x+y)/4)))
        sage: sorted(pts)
        [[2/3, 2], [3/4, 3/2], [1, 1]]

    """
    _, H, variable_map = _prepare_symbolic_fraction(F)
    variables = list(variable_map.values())
    if whitney_strat is not None:
        whitney_strat = [
            [
                tuple(SR(gen).subs(variable_map) for gen in component)
                for component in stratum
            ]
            for stratum in whitney_strat
        ]
    if linear_form is not None:
        linear_form = SR(linear_form).subs(variable_map)
    (
        expanded_R,
        vs,
        (lambda_, u_),
        r,
        r_variable_values,
    ) = _prepare_expanded_polynomial_ring(variables, direction=r, include_t=False)
    H = expanded_R(H)

    vsT = vs + list(r_variable_values.keys()) + [lambda_]

    # Compute the critical point system for each stratum
    pure_H = PolynomialRing(QQ, vs)

    if whitney_strat is None:
        whitney_strat = WhitneyStrat(Ideal(pure_H(H)), pure_H)
    else:
        # Cast symbolic generators for provided stratification into the correct ring
        whitney_strat = [
            prod([Ideal([pure_H(f) for f in comp]) for comp in stratum])
            for stratum in whitney_strat
        ]

    critical_points = []
    for d, stratum in enumerate(whitney_strat):
        for P_comp in PrimaryDecomposition(stratum):
            c = len(vs) - d
            P_ext = P_comp.change_ring(expanded_R)
            M = matrix([[v * f.derivative(v) for v in vs] for f in P_ext.gens()] + [r])
            # Add in min polys for the direction variables
            r_polys = [
                direction_value.minpoly().subs(direction_var)
                for direction_var, direction_value in r_variable_values.items()
            ]
            # Create ideal of expanded_R containing extended critical point equations
            ideal = P_ext + Ideal(
                M.minors(c + 1) + [(prod(vs) * lambda_ - 1)] + r_polys
            )
            # Saturate cpid by lower dimension stratum, if d > 0
            if d > 0:
                ideal = Saturate(ideal, whitney_strat[d - 1].change_ring(expanded_R))

            if ideal.dimension() < 0:
                continue
            P, Qs = _kronecker_representation(ideal.gens(), u_, vsT, linear_form)

            Pd = P.derivative()

            # Characterize all complex critical points in each stratum
            for u in P.roots(QQbar, multiplicities=False):
                rv = {
                    ri: (q / Pd).subs(u_=u)
                    for (ri, q) in zip(r_variable_values, Qs[len(vs) : -2])
                }
                if any(
                    rv[r_var] != r_value for r_var, r_value in r_variable_values.items()
                ):
                    continue

                w = [
                    collapse_zero_part(QQbar((q / Pd).subs(u_=u)))
                    for q in Qs[: len(vs)]
                ]
                critical_points.append(w)

    return critical_points


def _prepare_symbolic_fraction(F):
    r"""Extract polynomial numerators and denomiators from a symbolic fraction,
    including the replacement of variable names.

    INPUT:

    * ``F`` -- A symbolic fraction
    """
    G, H = F.numerator(), F.denominator()
    original_variables = H.variables()
    if any(v not in original_variables for v in G.variables()):
        raise ValueError(f"Numerator {G} has variables not in denominator {H}")
    new_variables = list(SR.var("acsvvar", len(original_variables)))
    variable_map = {v: new_v for v, new_v in zip(original_variables, new_variables)}
    G = G.subs(variable_map)
    H = H.subs(variable_map)
    frac_gcd = gcd(G, H)
    return G / frac_gcd, H / frac_gcd, variable_map


def _prepare_expanded_polynomial_ring(variables, direction=None, include_t=True):
    r"""Prepare an auxiliary polynomial ring for the diagonal asymptotics.

    INPUT:

    * ``variables`` -- variables in the rational function `F`
    * ``direction`` -- direction vector `r` for the asymptotics,
      defaults to the diagonal (all ones).
    * ``include_t`` -- whether to include the auxiliary variable `t`
      in the expanded ring.
    """
    # if direction r is not given, default to the diagonal
    if direction is None:
        direction = [1 for _ in variables]

    replaced_direction = copy(direction)

    # in case there are irrational numbers in the direction vector,
    # replace them with polynomial variables
    direction_variable_values = {}
    for idx, dir_entry in enumerate(replaced_direction):
        if AA(dir_entry).minpoly().degree() > 1:
            dir_var = SR.var(f"r{idx}")
            direction_variable_values[dir_var] = AA(dir_entry)
            replaced_direction[idx] = dir_var

    # auxiliary variables
    auxiliary_variables = []
    if include_t:
        auxiliary_variables.append(SR.var("t"))
    auxiliary_variables.append(SR.var("lambda_"))
    auxiliary_variables.append(SR.var("u_"))

    # create the expanded polynomial ring
    expanded_ring = PolynomialRing(
        QQ, list(variables) + list(direction_variable_values) + auxiliary_variables
    )
    variables = [expanded_ring(v) for v in variables]
    auxiliary_variables = [expanded_ring(v) for v in auxiliary_variables]
    replaced_direction = [expanded_ring(ri) for ri in replaced_direction]
    direction_variable_values = {
        expanded_ring(ri): val for (ri, val) in direction_variable_values.items()
    }
    return (
        expanded_ring,
        variables,
        auxiliary_variables,
        replaced_direction,
        direction_variable_values,
    )



# TODO - this should be removed once we have the global configuration settings
from sage.all import PolynomialRing, QQ, Hom
from sage_acsv.settings import ACSVSettings


def _construct_m2_morphims(ideal):
    R = ideal.gens()[0].parent()
    n = len(R.gens())
    RM2 = PolynomialRing(QQ, n, [f"x{i}" for i in range(n)])
    vs = list(RM2.gens())
    mor = Hom(R, RM2)(vs)
    inv = mor.inverse()
    return mor, inv


def PrimaryDecomposition(ideal):
    """Return the primary decomposition of an ideal.

    If a Macaulay2 interface is provided, it will be used instead
    of Sage's default implementation.

    INPUT:

    * ``ideal`` - A polynomial ideal
    """
    if ACSVSettings.get_default_groebner_backend() == ACSVSettings.Groebner.MACAULAY2:
        mor, inv = _construct_m2_morphims(ideal)
        ideal = ideal.apply_morphism(mor)
        m2 = ACSVSettings.get_macaulay2()
        return iter(J.sage().apply_morphism(inv) for J in m2.decompose(ideal))
    return ideal.primary_decomposition()


def Saturate(ideal_I, ideal_J):
    """Return the saturation of ideal I with respect to ideal J.

    INPUT:

    * ``ideal_I`` - A polynomial ideal
    * ``ideal_J`` - A polynomial ideal
    """
    return ideal_I.saturation(ideal_J)[0]


def GroebnerBasis(ideal):
    """Return a Groebner Basis of an ideal.

    If a Macaulay2 interface is provided, it will be used instead
    of Sage's default implementation.

    INPUT:

    * ``ideal`` - A polynomial ideal
    """
    if ACSVSettings.get_default_groebner_backend() == ACSVSettings.Groebner.MACAULAY2:
        mor, inv = _construct_m2_morphims(ideal)
        ideal = ideal.apply_morphism(mor)
        m2 = ACSVSettings.get_macaulay2()
        # This is really roundabout but for some reason returning the generators from M2 to sage
        # puts it in the wrong ring
        return m2.ideal(m2.gb(ideal).generators()).sage().apply_morphism(inv).gens()
    return ideal.groebner_basis()


def Radical(ideal):
    """Return the radical of an ideal.

    If a Macaulay2 interface is provided, it will be used instead
    of Sage's default implementation.

    INPUT:

    * ``ideal`` - A polynomial ideal
    """
    if ACSVSettings.get_default_groebner_backend() == ACSVSettings.Groebner.MACAULAY2:
        mor, inv = _construct_m2_morphims(ideal)
        ideal = ideal.apply_morphism(mor)
        m2 = ACSVSettings.get_macaulay2()
        return m2.radical(ideal).sage().apply_morphism(inv)
    return ideal.radical()



r"""A SageMath package for Analytic Combinatorics in Several Variables

This package provides tools for the study of the asymptotic behavior of
the sequence of coefficients of rational generating functions in
several variables along a given direction.

The public interface of our toolbox is provided by the following
functions and classes:

- :func:`.diagonal_asy` -- the central function of the package,
- :func:`.get_expansion_terms` -- helper function for extracting
  the terms in the output expansion with more structure,
- :func:`.ContributingCombinatorial` -- computes all contributing
  points of a combinatorial multivariate rational function,
- :func:`.MinimalCriticalCombinatorial` -- computes all non-zero minimal
  critical points of a combinatorial multivariate rational function,
- :func:`.CriticalPoints` -- compute all critical points of a combinatorial
  multivariate rational function,
- :class:`.ACSVSettings` -- a class for managing several package-global
  settings (like the default output format for :func:`.diagonal_asy`,
  or the backend used for Gröbner basis computations).

The following exmples illustrate some typical use cases. We
first import relevant functions and define the required
symbolic variables::

    sage: from sage_acsv import diagonal_asy, get_expansion_terms
    sage: var('w x y z')
    (w, x, y, z)

The asymptotic expansion for the sequence of central binomial
coefficients, `\binom{2n}{n}`, generated along the `(1, 1)`-diagonal
of `F(x, y) = \frac{1}{1 - x - y}` is simply computed as
::

    sage: diagonal_asy(1 / (1 - x - y))
    1/sqrt(pi)*4^n*n^(-1/2) + O(4^n*n^(-3/2))

The precision of the expansion can be controlled by the
``expansion_precision`` keyword argument::

    sage: diagonal_asy(1 / (1 - x - y), expansion_precision=4)
    1/sqrt(pi)*4^n*n^(-1/2) - 1/8/sqrt(pi)*4^n*n^(-3/2) + 1/128/sqrt(pi)*4^n*n^(-5/2) + 5/1024/sqrt(pi)*4^n*n^(-7/2) + O(4^n*n^(-9/2))

::

    sage: F1 = (2*y^2 - x)/(x + y - 1)
    sage: diagonal_asy(F1, expansion_precision=3)
    1/4/sqrt(pi)*4^n*n^(-3/2) + 3/32/sqrt(pi)*4^n*n^(-5/2) + O(4^n*n^(-7/2))

::

    sage: F2 = (1+x)*(1+y)/(1-w*x*y*(x+y+1/x+1/y))
    sage: diagonal_asy(F2, expansion_precision=3)
    4/pi*4^n*n^(-1) - 6/pi*4^n*n^(-2) + 1/pi*4^n*n^(-3)*(e^(I*arg(-1)))^n + 19/2/pi*4^n*n^(-3) + O(4^n*n^(-4))

The following example is from Apéry's proof concerning the
irrationality of `\zeta(3)`::

    sage: F3 = 1/(1 - w*(1 + x)*(1 + y)*(1 + z)*(x*y*z + y*z + y + z + 1))
    sage: apery_expansion = diagonal_asy(F3, expansion_precision=2); apery_expansion
    1.225275868941647?/pi^(3/2)*33.97056274847714?^n*n^(-3/2) - 0.5128314911970734?/pi^(3/2)*33.97056274847714?^n*n^(-5/2) + O(33.97056274847714?^n*n^(-7/2))

While the representation might suggest otherwise, the numerical
constants in the expansion are not approximations, but in fact
explicitly known algebraic numbers. We can use the
:func:`.get_expansion_terms` function to inspect them closer::

    sage: coefs = get_expansion_terms(apery_expansion); coefs
    [Term(coefficient=1.225275868941647?, pi_factor=pi^(-3/2), base=33.97056274847714?, power=-3/2),
     Term(coefficient=-0.5128314911970734?, pi_factor=pi^(-3/2), base=33.97056274847714?, power=-5/2)]
    sage: coefs[0].coefficient.radical_expression()
    1/4*sqrt(17/2*sqrt(2) + 12)
    sage: coefs[0].base.radical_expression()
    12*sqrt(2) + 17

::

    sage: F4 = -1/(1 - (1 - x - y)*(20 - x - 40*y))
    sage: diagonal_asy(F4, expansion_precision=2)
    0.09677555757474702?/sqrt(pi)*5.884442204019508?^n*n^(-1/2) + 0.002581950724843528?/sqrt(pi)*5.884442204019508?^n*n^(-3/2) + O(5.884442204019508?^n*n^(-5/2))

The package raises an exception if it detects that some of the
requirements are not met::

    sage: F5 = 1/(x^4*y + x^3*y + x^2*y + x*y - 1)
    sage: diagonal_asy(F5)
    Traceback (most recent call last):
    ...
    ACSVException: No smooth minimal critical points found.

::

    sage: F6 = 1/((-x + 1)^4 - x*y*(x^3 + x^2*y - x^2 - x + 1))
    sage: diagonal_asy(F6)  # long time
    Traceback (most recent call last):
    ...
    ACSVException: No contributing points found.

Here is the asymptotic growth of the Delannoy numbers::

    sage: F7 = 1/(1 - x - y - x*y)
    sage: diagonal_asy(F7)
    1.015051765128218?/sqrt(pi)*5.828427124746190?^n*n^(-1/2) + O(5.828427124746190?^n*n^(-3/2))

::

    sage: F8 = 1/(1 - x^7)
    sage: diagonal_asy(F8)
    1/7 + 1/7*(e^(I*arg(-0.2225209339563144? + 0.9749279121818236?*I)))^n + 1/7*(e^(I*arg(-0.2225209339563144? - 0.9749279121818236?*I)))^n + 1/7*(e^(I*arg(-0.9009688679024191? + 0.4338837391175581?*I)))^n + 1/7*(e^(I*arg(-0.9009688679024191? - 0.4338837391175581?*I)))^n + 1/7*(e^(I*arg(0.6234898018587335? + 0.7818314824680299?*I)))^n + 1/7*(e^(I*arg(0.6234898018587335? - 0.7818314824680299?*I)))^n + O(n^(-1))

This example is for a generating function whose singularities have
very close moduli::

    sage: F9 = 1/(8 - 17*x^3 - 9*x^2 + 7*x)
    sage: diagonal_asy(F9, return_points=True)
    (0.03396226416457560?*1.285654384750451?^n + O(1.285654384750451?^n*n^(-1)),
     [[0.7778140158516262?]])

"""

from sage_acsv.asymptotics import *
from sage_acsv.kronecker import *
from sage_acsv.helpers import get_expansion_terms
from sage_acsv.settings import ACSVSettings



import time
import logging

logging.basicConfig()
acsv_logger = logging.getLogger("sage_acsv")


class Timer:
    def __init__(self):
        self._time_created = time.monotonic()
        self._last_checkpoint = self._time_created

    def checkpoint(self, name=None):
        if name:
            time_diff = time.monotonic() - self._last_checkpoint
            acsv_logger.info(
                f"{bcolors.OKBLUE}Timer:{bcolors.ENDC} "
                f"Executed {name} in {time_diff} seconds."
            )

        self._last_checkpoint = time.monotonic()

    def total(self, name):
        if self.show_time and name:
            time_diff = time.monotonic() - self._time_created
            acsv_logger.info(
                f"{bcolors.OKBLUE}Timer:{bcolors.ENDC}Total runtime "
                f"of {name} is {time_diff} seconds."
            )


class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    ERROR = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"



from sage.all import Ideal, PolynomialRing, ProjectiveSpace, QQ
from sage.all import Combinations, matrix

from sage_acsv.macaulay2 import PrimaryDecomposition, GroebnerBasis, Saturate, Radical


def Con(X, P, RZ):
    r"""Compute the ideal associated with the map sending `X` to its conormal space.

    The map is also denoted as `\operatorname{Con}(X)`.
    """
    vs = P.gens()
    vzs = tuple(v for v in RZ.gens() if v not in vs)
    c = X.codimension()

    X = P.subscheme(Radical(X.defining_ideal()))
    M = X.Jacobian_matrix()
    Jac = X.Jacobian()
    JacZ = Ideal(matrix([list(vzs)] + list(M)).minors(c + 1))

    return Saturate(X.defining_ideal().change_ring(RZ) + JacZ, Jac.change_ring(RZ))


def Decompose(Y, X, P, R, RZ):
    r"""Given varieties `X` and `Y`, return the points in `Y` that fail
    Whitney's condition `B` with respect to `X`.
    """
    d = Y.dimension()
    Ys = [P.subscheme(1) for _ in range(d + 1)]
    Ys[-1] = Y

    J = Con(X, P, RZ) + Y.defining_ideal().change_ring(RZ)
    J = Ideal(GroebnerBasis(J))
    for IQ in PrimaryDecomposition(J):
        K = IQ.elimination_ideal(
            [v for v in RZ.gens() if v not in R.gens()]
        ).change_ring(R)
        W = P.subscheme(K)
        if W.dimension() < Y.dimension():
            i = W.dimension()
            Ys[i] = Ys[i].union(W)
    return Ys


def merge_stratifications(Xs, Ys):
    r"""Merge two stratifications."""
    # Ensures Xs > Ys
    if len(Ys) > len(Xs):
        return merge_stratifications(Ys, Xs)

    res = []
    for i in range(len(Ys)):
        res.append(Xs[i].union(Ys[i]))
    for i in range(len(Ys), len(Xs)):
        res.append(Xs[i].union(Ys[-1]))

    return res


def WhitneyStratProjective(X, P):
    r"""
    Computes a WhitneyStratification of projective variety X in the ring P
    """
    X = P.subscheme(Radical(X.defining_ideal()))
    vs = P.gens()
    k = X.dimension()

    R = P.gens()[0].parent()
    RZ = PolynomialRing(R, ["t%d" % i for i in range(len(vs))])
    RZ = RZ.flattening_morphism()(vs[0]).parent()

    Xs = [P.subscheme(1) for _ in range(k + 1)]
    Xs[-1] = X
    X_sing = X.intersection(P.subscheme(X.Jacobian()))
    mu = X_sing.dimension()

    for IZ in PrimaryDecomposition(X_sing.defining_ideal()):
        Z = P.subscheme(IZ)
        i = Z.dimension()
        Xs[i] = Xs[i].union(Z)

    for d in range(mu + 1)[::-1]:
        Xd = Xs[d]
        if Xd.dimension() < d:
            continue
        Xs = merge_stratifications(Xs, Decompose(Xd, X, P, R, RZ))
        Xs = merge_stratifications(Xs, WhitneyStratProjective(Xd, P))

    return Xs


def WhitneyStrat(IX, R):
    r"""Computes the Whitney Stratification of a pure-dimensional algebraic variety.

    Uses an algorithm developed by Helmer and Nanda (2022).

    INPUT:

    * ``IX`` -- A `k`-dimensional polynomial ideal representation of the algebraic variety `X`
    * ``R`` -- Base ring of the ideal. Should be a PolynomialRing object

    OUTPUT:

    A list ``[IX_0, IX_1, ..., IX_k]`` of polynomial ideals representing the Whitney stratification of `X`.
    ``IX_j`` reprensents the `j`-dimensional stratum.

    EXAMPLES::

        sage: from sage_acsv.whitney import WhitneyStrat
        sage: R.<x,y,z> = PolynomialRing(QQ, 3)
        sage: WhitneyStrat(Ideal(y^2+x^3-y^2*z^2), R)
        [Ideal (y, x, z^2 - 1) of Multivariate Polynomial Ring in x, y, z over Rational Field,
         Ideal (y, x) of Multivariate Polynomial Ring in x, y, z over Rational Field,
         Ideal (y^2*z^2 - x^3 - y^2) of Multivariate Polynomial Ring in x, y, z over Rational Field]
    """
    vs = R.gens()
    d = len(vs)
    # Check if IX = V(H) and H factors smoothly -
    # if so, the whitney strat is just the intersection of subsets of the components
    if len(IX.gens()) == 1:
        factors = [fm[0] for fm in IX.gens()[0].factor()]
        if all(
            Ideal(
                fs + matrix([[f.derivative(v) for v in vs] for f in fs]).minors(len(fs))
            ).dimension()
            < 0
            for fs in Combinations(factors)
            if len(fs) >= 1
        ):
            strat = [Ideal(R(1)) for _ in range(d)]
            strat[-1] = Ideal(IX.gens())
            for k in reversed(range(1, d)):
                Jac = matrix([[f.derivative(v) for v in vs] for f in strat[k].gens()])
                sing = Radical(Ideal(Jac.minors(d - k) + strat[k].gens()))
                for i in range(max(sing.dimension(), 0), k):
                    strat[i] = Radical(sing.intersection(strat[i]))
            return strat

    P, vsP = ProjectiveSpace(QQ, len(vs), list(vs) + ["z0"]).objgens()

    z0 = vsP[-1]
    R_hom = z0.parent()
    proj_strat = WhitneyStratProjective(
        P.subscheme(IX.change_ring(R_hom).homogenize(z0)), P
    )

    strat = [Ideal(R(1)) for _ in range(len(proj_strat))]
    for stratum in proj_strat:
        for Id in PrimaryDecomposition(stratum.defining_ideal()):
            newId = Id.subs({z0: 1}).change_ring(R)
            for k in range(newId.dimension(), len(strat)):
                strat[k] = Radical(strat[k].intersection(newId))

    return strat



from __future__ import annotations

from dataclasses import dataclass

from sage.groups.misc_gps.argument_groups import ArgumentByElementGroup
from sage.rings.asymptotic.asymptotic_ring import AsymptoticRing, AsymptoticExpansion
from sage.rings.asymptotic.growth_group import (
    ExponentialGrowthGroup,
    MonomialGrowthGroup,
)
from sage.rings.qqbar import AlgebraicNumber, QQbar
from sage.symbolic.expression import Expression
from sage.symbolic.ring import SymbolicRing
from sage.symbolic.operators import add_vararg

from sage.all import (
    AA,
    QQ,
    SR,
    Ideal,
    Polyhedron,
    ceil,
    gcd,
    matrix,
    randint,
    vector,
    kronecker_delta,
    prod,
    pi,
)


@dataclass
class Term:
    r"""A dataclass for storing the decomposed terms of an asymptotic expression.

    INPUT:

    * ``coefficient`` -- The coefficient of the term
    * ``pi_factor`` -- The factor of pi in the term
    * ``base`` -- The base of the term
    * ``power`` -- The power of the term

    OUTPUT:

    A dataclass with the given attributes.

    EXAMPLES::

        sage: from sage_acsv.helpers import Term
        sage: Term(1, 1/sqrt(pi), 4, -1/2)
        Term(coefficient=1, pi_factor=1/sqrt(pi), base=4, power=-1/2)
    """

    coefficient: Expression | AlgebraicNumber
    pi_factor: Expression
    base: Expression | AlgebraicNumber
    power: Expression | AlgebraicNumber

    def __lt__(self, other):
        return (self.base, self.power) < (other.base, other.power)


def collapse_zero_part(algebraic_number: AlgebraicNumber) -> AlgebraicNumber:
    if algebraic_number.real().is_zero():
        algebraic_number = QQbar(algebraic_number.imag()) * QQbar(-1).sqrt()
    if algebraic_number.imag().is_zero():
        algebraic_number = QQbar(algebraic_number.real())
    algebraic_number.simplify()
    return algebraic_number


def RationalFunctionReduce(G, H):
    r"""Reduction of G and H by dividing out their GCD.

    INPUT:

    * ``G``, ``H`` -- polynomials

    OUTPUT:

    A tuple ``(G/d, H/d)``, where ``d`` is the GCD of ``G`` and ``H``.
    """
    g = gcd(G, H)
    return G / g, H / g


def GenerateLinearForm(system, vsT, u_, linear_form=None):
    r"""Generate a linear form of the input system.

    This is an integer linear combination of the variables that
    take unique values on the solutions of the system. The
    generated linear form is with high probability separating.

    INPUT:

    * ``system`` -- A polynomial system of equations
    * ``vsT`` -- A list of variables in the system
    * ``u_`` -- A variable not in the system
    * ``linear_form`` -- (Optional) A precomputed linear form in the
      variables of the system. If passed, the returned form is
      based on the given linear form and not randomly generated.

    OUTPUT:

    A linear form.
    """
    if linear_form is not None:
        return u_ - linear_form

    maxcoeff = ceil(
        max([max([abs(x) for x in f.coefficients()]) for f in system if f != 0])
    )
    maxdegree = max([f.degree() for f in system])
    return u_ - sum(
        [
            randint(-maxcoeff * maxdegree - 31, maxcoeff * maxdegree + 31) * z
            for z in vsT
        ]
    )


def GetHessian(H, variables, r, critical_point=None):
    r"""Computes the Hessian of a given map.

    The map underlying `Hess` is defined as

    .. math::

        (z_1, \ldots, z_{d-1}) \mapsto z_1 \cdots z_{d-1} \log(g(z_1, \ldots, z_{d-1})),

    with `g` determined via the Implicit Function Theorem for the equation

    .. math::

        H(z_1,...,z_{d-1}, g(z_1,...,z_{d-1})) = 0`

    at a critical point in direction `r`.

    INPUT:

    * ``H`` -- A polynomial; the denominator of the rational generating function
      `F = G/H`.
    * ``vs`` -- A list of variables ``z_1, ..., z_d``
    * ``r`` -- The direction. A vector of length `d` with positive algebraic numbers
      (usually integers) as coordinates.
    * ``critical_point`` -- An optional critical point where the Hessian should be evaluated at.
      If not specified, the symbolic Hessian is returned.

    OUTPUT:

    A matrix representing the Hessian of the given map.
    """
    z_d = variables[-1]
    d = len(variables)

    # Build d x d matrix of U[i,j] = z_i * z_j * H'_{z_i * z_j}
    U = matrix(
        [
            [
                v1 * v2 * H.derivative(v1, v2) / (z_d * H.derivative(z_d))
                for v2 in variables
            ]
            for v1 in variables
        ]
    )

    try:
        V = [QQ(r[k] / r[-1]) for k in range(d)]
    except (ValueError, TypeError):
        V = [AA(r[k] / r[-1]) for k in range(d)]

    # Build (d-1) x (d-1) Matrix for Hessian
    hessian = [
        [
            V[i] * V[j]
            + U[i][j]
            - V[j] * U[i][-1]
            - V[i] * U[j][-1]
            + V[i] * V[j] * U[-1][-1]
            for j in range(d - 1)
        ]
        for i in range(d - 1)
    ]
    for i in range(d - 1):
        hessian[i][i] = hessian[i][i] + V[i]

    hessian = matrix(hessian)
    if critical_point is not None:
        hessian = hessian.subs(critical_point)
    return hessian


def NewtonSeries(phi, variables, series_precision):
    r"""Computes the series expansion of an implicitly defined function.

    The function `g(x)` for which a series expansion is computed satisfies

    .. math::

        \Phi(x, g(x)) = 0

    INPUT:

    * ``phi`` -- A polynomial; the equation defining the function that is expanded.
    * ``variables`` -- A list of variables in the equation. The last variable in this
      list is the variable corresponding to `g(x)`.
    * ``series_precision`` -- A positive integer, the precision of the series expansion.

    OUTPUT:

    A series expansion of the function `g(x)`.

    EXAMPLES::

        sage: from sage_acsv.helpers import NewtonSeries
        sage: R.<x, T> = QQ[]
        sage: NewtonSeries(x*T^2 - T + 1, [x, T], 7)
        132*x^6 + 42*x^5 + 14*x^4 + 5*x^3 + 2*x^2 + x + 1

    """

    X = variables[:-1]
    Y = variables[-1]

    def ModX(F, N):
        return F.mod(Ideal(X) ** N)

    def ModY(F, N):
        return F.mod(Y**N)

    def Mod(F, N):
        return ModX(ModY(F, N), N)

    def NewtonRecur(H, N):
        if N == 1:
            return 0, 1 / H.derivative(Y).subs({v: 0 for v in variables})
        F, G = NewtonRecur(H, ceil(N / 2))
        G = G + (1 - G * H.derivative(Y).subs({Y: F})) * G
        F = F - G * H.subs({Y: F})
        return ModX(F, N), ModX(G, ceil(N / 2))

    return NewtonRecur(Mod(phi, series_precision), series_precision)[0]


def ImplicitHessian(Hs, vs, r, subs):
    r"""Compute the Hessian of an implicitly defined function.

    Given a transverse intersection point `w` in `H_1(w),\dots,H_s(w)=0`, we can parametrize `V(H_1,\dots,H_s)`
    near `w` by writing `z_{d-s+j} = g_j(z_1,\dots,z_{d-s})`.

    Let `h(\theta_1,\dots,\theta_{d-s}) = \sum_{j=1}^s r_{d-s+j}\log g_j({w_1 exp(i\theta_1) \dots w_{d-s} exp(i\theta_{d-s})})`.
    This function returns the Hessian of h.

    INPUT:

    * ``Hs`` -- A list of polynomials `H`
    * ``vs`` -- A list of variables in the equation
    * ``r`` -- A direction vector
    * ``subs`` -- a dic `{v_i:w_i}` for point w

    OUTPUT:

    The Hessian of the implicitly defined function `h` defined above.

    EXAMPLES::

        sage: from sage_acsv.helpers import ImplicitHessian
        sage: R.<x,y,z,w> = PolynomialRing(QQ,4)
        sage: Hs = [
        ....:     z^2+z*w+x*y-4,
        ....:     w^3+z*x-y
        ....: ]
        sage: ImplicitHessian(Hs, [x,y,z,w], [1,1,1,1], {x:1,y:1,z:1,w:1})
        [21/32     0]
        [    0   7/8]
    """

    d = len(vs)
    s = len(Hs)
    Hs, vs = [SR(H) for H in Hs], [SR(v) for v in vs]
    if subs:
        subs = {SR(v): val for v, val in subs.items()}
    dHdg = matrix([[H.derivative(v) for v in vs[d - s :]] for H in Hs])
    dHdv = matrix([[H.derivative(v) for v in vs[: d - s]] for H in Hs])
    dgdv = -dHdg.inverse() * dHdv

    d2gdv2 = [
        [
            -dHdg.inverse()
            * (
                vector([H.derivative(vs[i]).derivative(vs[j]) for H in Hs])
                + matrix(
                    [
                        [H.derivative(vs[i]).derivative(g) for g in vs[d - s :]]
                        for H in Hs
                    ]
                )
                * dgdv.column(j)
                + matrix(
                    [
                        [H.derivative(g).derivative(vs[j]) for g in vs[d - s :]]
                        for H in Hs
                    ]
                )
                * dgdv.column(i)
                + vector(
                    [
                        matrix(
                            [
                                [H.derivative(g1).derivative(g2) for g1 in vs[d - s :]]
                                for g2 in vs[d - s :]
                            ]
                        )
                        * dgdv.column(i)
                        * dgdv.column(j)
                        for H in Hs
                    ]
                )
            )
            for i in range(d - s)
        ]
        for j in range(d - s)
    ]

    Hess = matrix(
        [
            [
                sum(
                    r[k]
                    * (
                        -vs[i] * vs[j] * d2gdv2[i][j][k - (d - s)] * vs[k]
                        - kronecker_delta(i, j) * dgdv[k - (d - s), j] * vs[k] * vs[i]
                        + vs[i] * vs[j] * dgdv[k - (d - s), i] * dgdv[k - (d - s), j]
                    )
                    / vs[k] ** 2
                    for k in range(d - s, d)
                )
                for i in range(d - s)
            ]
            for j in range(d - s)
        ]
    )
    if subs:
        return Hess.subs(subs)

    return Hess


def IsContributing(vs, pt, r, factors, c):
    r"""Determines if minimal critical point `pt` where singular variety has transverse square-free factorization
    is contributing; that is, whether `r` is in the interior
    of the scaled log-normal cone of `factors` at `pt`

    INPUT:

    * ``vs`` -- A list of variables
    * ``pt`` -- A point
    * ``r`` -- A direction vector
    * ``factors`` -- A list of factors of `H` for which `pt` vanishes
    * ``c`` -- The co-dimension of the intersection of factors

    OUTPUT:

    If vs is contributing

    EXAMPLES::

        sage: from sage_acsv.helpers import IsContributing
        sage: R.<x,y> = PolynomialRing(QQ, 2)
        sage: IsContributing([x,y], [1,1], [17/24, 7/24], [1-(2*x+y)/3,1-(3*x+y)/4], 2)
        True
        sage: IsContributing([x,y], [1,1], [1, 1], [1-(2*x+y)/3,1-(3*x+y)/4], 2)
        False

    """
    critical_subs = {v: point for v, point in zip(vs, pt)}
    # Compute irreducible components of H that contain the point
    vanishing_factors = list(f for f in factors if f.subs(critical_subs) == 0)
    for f in vanishing_factors:
        if all([f.derivative(v).subs(critical_subs) == 0 for v in vs]):
            raise ACSVException(
                f"Critical point {pt} lies in non-smooth part of component {f}"
            )

    vkjs = []
    for f in vanishing_factors:
        for v in vs:
            if f.derivative(v).subs(critical_subs) != 0:
                vkjs.append(v)
                break
        else:
            # In theory this shouldn't ever happen, now that we check the condition before
            raise ACSVException(
                f"All partials of component {vanishing_factors} vanish at point {pt}"
            )

    normals = matrix(
        list(
            [
                AA(
                    f.derivative(v).subs(critical_subs)
                    * critical_subs[v]
                    / (critical_subs[vkj] * f.derivative(vkj).subs(critical_subs))
                )
                for v in vs
            ]
            for vkj, f in zip(vkjs, vanishing_factors)
        )
    )

    polytope = Polyhedron(rays=normals)
    if r not in polytope:
        return False
    elif any([r in f for f in polytope.faces(c - 1)]):
        # If r is in the boundary of the log normal cone, point is non-generic
        raise ACSVException(
            f"Non-generic critical point found - {pt} is contained in {len(vs) - c}-dimensional stratum"
        )
    return True


def get_expansion_terms(
    expr: tuple | list[tuple] | Expression | AsymptoticExpansion,
) -> list[Term]:
    r"""Determines coefficients for each n^k that appears in the asymptotic expression.

    INPUT:

    * ``expr`` -- An asymptotic expression, symbolic expression, ACSV tuple, or list of ACSV tuples

    OUTPUT:

    A list of :class:`.DecomposedTerm` objects (with attributes ``coefficient``, ``pi_factor``,
    ``base`` and ``power``), each representing a summand in the fully expanded expression.

    EXAMPLES::

        sage: from sage_acsv import diagonal_asy, get_expansion_terms
        sage: var('x y z')
        (x, y, z)
        sage: res = diagonal_asy(1/(1 - x - y), r=[1,1], expansion_precision=2)
        sage: coefs = sorted(get_expansion_terms(res), reverse=True)
        sage: coefs
        [Term(coefficient=1, pi_factor=1/sqrt(pi), base=4, power=-1/2),
         Term(coefficient=-1/8, pi_factor=1/sqrt(pi), base=4, power=-3/2)]
        sage: res = diagonal_asy(1/(1 - x - y), r=[1,1], expansion_precision=2, output_format="tuple")
        sage: sorted(get_expansion_terms(res)) == sorted(coefs)
        True
        sage: res = diagonal_asy(1/(1 - x - y), r=[1,1], expansion_precision=2, output_format="symbolic")
        sage: sorted(get_expansion_terms(res)) == sorted(coefs)
        True

    ::

        sage: res = diagonal_asy(1/(1 - x^7))
        sage: get_expansion_terms(res)
        [Term(coefficient=1/7, pi_factor=1, base=0.6234898018587335? + 0.7818314824680299?*I, power=0),
         Term(coefficient=1/7, pi_factor=1, base=0.6234898018587335? - 0.7818314824680299?*I, power=0),
         Term(coefficient=1/7, pi_factor=1, base=-0.2225209339563144? + 0.9749279121818236?*I, power=0),
         Term(coefficient=1/7, pi_factor=1, base=-0.2225209339563144? - 0.9749279121818236?*I, power=0),
         Term(coefficient=1/7, pi_factor=1, base=-0.9009688679024191? + 0.4338837391175582?*I, power=0),
         Term(coefficient=1/7, pi_factor=1, base=-0.9009688679024191? - 0.4338837391175582?*I, power=0),
         Term(coefficient=1/7, pi_factor=1, base=1, power=0)]

    ::

        sage: res = diagonal_asy(1/(1 - x - y^2))
        sage: coefs = get_expansion_terms(res); coefs
        [Term(coefficient=0.6123724356957945?, pi_factor=1/sqrt(pi), base=-2.598076211353316?, power=-1/2),
         Term(coefficient=0.6123724356957945?, pi_factor=1/sqrt(pi), base=2.598076211353316?, power=-1/2)]
        sage: coefs[0].coefficient.parent()
        Algebraic Field
        sage: coefs[0].coefficient.radical_expression()
        1/2*sqrt(3/2)

    ::

        sage: F2 = (1+x)*(1+y)/(1-z*x*y*(x+y+1/x+1/y))
        sage: res = diagonal_asy(F2, expansion_precision=3)
        sage: coefs = get_expansion_terms(res); coefs
        [Term(coefficient=4, pi_factor=1/pi, base=4, power=-1),
         Term(coefficient=1, pi_factor=1/pi, base=-4, power=-3),
         Term(coefficient=-6, pi_factor=1/pi, base=4, power=-2),
         Term(coefficient=19/2, pi_factor=1/pi, base=4, power=-3)]

    ::

        sage: res = diagonal_asy(3/(1 - x))
        sage: get_expansion_terms(res)
        [Term(coefficient=3, pi_factor=1, base=1, power=0)]

    ::

        sage: res = diagonal_asy((x - y)/(1 - x - y))
        sage: get_expansion_terms(res)
        []

    """
    n = SR.var("n")
    if isinstance(expr, tuple):
        expr = SR(expr[0] ** n * prod(expr[1:]))
    elif isinstance(expr, list):
        expr = SR(sum([tup[0] ** n * prod(tup[1:]) for tup in expr]))
    elif isinstance(expr.parent(), AsymptoticRing):
        expr = expr.exact_part()
        symbolic_expr = SR.zero()
        for summand in expr.summands:
            symbolic_summand = summand.coefficient
            for factor in summand.growth.value:
                if isinstance(factor.parent(), MonomialGrowthGroup):
                    symbolic_summand *= n**factor.exponent
                elif isinstance(factor.parent(), ExponentialGrowthGroup):
                    if isinstance(factor.base.parent(), ArgumentByElementGroup):
                        symbolic_summand *= factor.base._element_**n
                    else:
                        symbolic_summand *= factor.base**n
            symbolic_expr += symbolic_summand
        expr = symbolic_expr

    if not isinstance(expr.parent(), type(SR)):
        raise ACSVException(f"Cannot deal with expression of type {expr.parent()}")

    if len(expr.args()) > 1:
        raise ACSVException("Cannot process multivariate symbolic expression.")

    # If expression is the sum of a bunch of terms, handle each one separately
    expr = expr.expand()
    if expr.is_zero():
        return []

    terms = [expr]
    if expr.operator() == add_vararg:
        terms = expr.operands()

    decomposed_terms = []
    for summand in terms:
        term = Term(
            coefficient=QQ.one(), pi_factor=QQ.one(), base=QQ.one(), power=QQ.zero()
        )
        if summand in QQbar:
            term.coefficient = QQbar(summand)
            decomposed_terms.append(term)
            continue

        for v in summand.operands():
            if n in v.args():
                if v.degree(n) != 0:
                    term.power += v.degree(n)
                else:
                    term.base *= v.operands()[0]
            elif v.degree(pi) != 0:
                pi_deg = v.degree(pi)
                term.pi_factor *= pi**pi_deg
                term.coefficient *= v.coefficient(pi**pi_deg)
            else:
                term.coefficient *= v

        for attr in ("coefficient", "base", "power"):
            elem = getattr(term, attr)
            if isinstance(elem.parent(), SymbolicRing) and elem in QQbar:
                setattr(term, attr, QQbar(elem))

        decomposed_terms.append(term)

    return decomposed_terms


class ACSVException(Exception):
    def __init__(self, message, retry=False):
        super().__init__(message)
        self._message = message
        self._retry = retry

    def __str__(self):
        return self._message

    @property
    def retry(self):
        return self._retry



from sage.all import PolynomialRing, QQ, gcd
from sage.rings.polynomial.multi_polynomial_ideal import MPolynomialIdeal

from sage_acsv.helpers import ACSVException, GenerateLinearForm
from sage_acsv.debug import acsv_logger
from sage_acsv.msolve import get_parametrization
from sage_acsv.macaulay2 import GroebnerBasis, Radical
from sage_acsv.settings import ACSVSettings


def _kronecker_representation_sage(system, u_, vs, linear_form=None):
    r"""Computes the Kronecker Representation of a system of polynomials.

    This method is intended for internal use and requires a consistent
    setup of parameters. Use the :func:`.kronecker` wrapper function
    to avoid doing the setup yourself.

    INPUT:

    * ``system`` -- A system of polynomials in ``d`` variables
    * ``u_`` -- Variable not contained in the variables in system
    * ``vs`` -- Variables of the system
    * ``linear_form`` -- (Optional) A linear combination of the input
      variables that separates the critical point solutions

    OUTPUT:

    A polynomial ``P`` and ``d`` polynomials ``Q1, ..., Q_d`` such
    that ``z_i = Q_i(u)/P'(u)`` for ``u`` ranging over the roots of ``P``

    Examples::

        sage: from sage_acsv.kronecker import kronecker
        sage: var('x, y')
        (x, y)
        sage: kronecker(
        ....:     [x**3+y**3-10, y**2-2],
        ....:     [x, y],
        ....:     linear_form=x + y
        ....: )  # indirect doctest
        (u_^6 - 6*u_^4 - 20*u_^3 + 36*u_^2 - 120*u_ + 100,
         [60*u_^3 - 72*u_^2 + 360*u_ - 600, 12*u_^4 - 72*u_^2 + 240*u_])
    """

    # Generate a linear form
    linear_form = GenerateLinearForm(system, vs, u_, linear_form)

    expanded_R = u_.parent()

    rabinowitsch_R = PolynomialRing(
        QQ, list(expanded_R.gens()), len(expanded_R.gens()), order="degrevlex"
    )
    u_ = rabinowitsch_R(u_)

    rabinowitsch_system = [rabinowitsch_R(f) for f in system]
    rabinowitsch_system.append(rabinowitsch_R(linear_form))

    # Compute Grobner basis for ordered system of polynomials
    ideal = MPolynomialIdeal(rabinowitsch_R, rabinowitsch_system)
    try:
        ideal = MPolynomialIdeal(rabinowitsch_R, GroebnerBasis(ideal))
    except Exception:
        raise ACSVException(
            "Trouble computing Groebner basis. System may be too large."
        )

    if ideal.dimension() != 0:
        raise ACSVException(
            f"Ideal {ideal} is not 0-dimensional. Cannot compute Kronecker representation."
        )

    ideal = Radical(ideal)
    gb = ideal.transformed_basis("fglm")

    rabinowitsch_R = rabinowitsch_R.change_ring(order="lex")
    u_ = rabinowitsch_R(u_)

    Ps = [
        p
        for p in gb
        if len(p.variables()) != 0 and not any([z in vs for z in p.variables()])
    ]
    if len(Ps) != 1:
        acsv_logger.debug(
            f"Rabinowitsch system: {rabinowitsch_system}\nPs: {Ps}\nbasis: {gb}"
        )
        raise ACSVException(
            "No P polynomial found for Kronecker Representation.", retry=True
        )
    u_ = Ps[0].variables()[0]
    R = PolynomialRing(QQ, u_)
    P = R(Ps[0])
    P, _ = P.quo_rem(gcd(P, P.derivative(u_)))
    Pd = P.derivative(u_)

    # Find Q_i for each variable
    Qs = []
    for z in vs:
        z = rabinowitsch_R(z)
        eqns = [f for f in gb if z in f.variables()]
        if len(eqns) != 1:
            acsv_logger.debug(f"equations: {eqns}\nz: {z}\nvs: {vs}")
            raise ACSVException("Linear form does not separate the roots.", retry=True)

        eq = eqns[0].polynomial(z)

        if eq.degree() != 1:
            acsv_logger.debug(f"eq: {eq}\nz: {z}")
            raise ACSVException("Linear form does not separate the roots.", retry=True)
        _, rem = (Pd * eq.roots()[0][0]).quo_rem(P)
        Qs.append(rem)

    # Forget base ring, move to univariate polynomial ring in u over a field
    Qs = [R(Q) for Q in Qs]

    return P, Qs


def _kronecker_representation_msolve(system, u_, vs):
    result = get_parametrization(vs, system)
    _, nvars, _, msvars, _, param = result[1]

    # msolve may reorder the variables, so order them back
    Qparams = param[1][2]
    vsExt = [str(v) for v in vs]
    # Check if no new variable was created by msolve
    # If so, the linear_form used is just zd
    # i.e. u_ = zd and Qd = zd * P'(zd)
    if nvars == len(vs):
        Pdz = [0] + [-c for c in param[1][1][1]]
        Qparams.append([[1, Pdz], 1])
    pidx = [msvars.index(v) for v in vsExt]
    Qparams = [Qparams[i] for i in pidx]

    R = PolynomialRing(QQ, u_)
    u_ = R(u_)

    P_coeffs = param[1][0][1]
    P = sum([c * u_**i for (i, c) in enumerate(P_coeffs)])

    Qs = []
    for Q_param in Qparams:
        Q_coeffs = Q_param[0][1]
        c_div = Q_param[1]
        Q = -sum([c * u_**i for (i, c) in enumerate(Q_coeffs)]) / c_div
        Qs.append(Q)

    return P, Qs


def _kronecker_representation(system, u_, vs, linear_form=None):
    r"""Computes the Kronecker Representation of a system of polynomials

    Internal intermediate function for choosing the Kronecker backend implementation

    INPUT:

    * ``system`` -- A system of polynomials in ``d`` variables
    * ``u_`` -- Variable not contained in the variables in system
    * ``vs`` -- Variables of the system
    * ``linear_form`` -- (Optional) A linear combination of the
      input variables that separates the critical point solutions

    OUTPUT:

    A polynomial ``P`` and ``d`` polynomials ``Q1, ..., Q_d`` such that
    ``z_i = Q_i(u)/P'(u)`` for ``u`` ranging over the roots of ``P``.
    """
    if ACSVSettings.get_default_kronecker_backend() == ACSVSettings.Kronecker.MSOLVE:
        return _kronecker_representation_msolve(system, u_, vs)
    return _kronecker_representation_sage(system, u_, vs, linear_form=linear_form)


def kronecker(system, vs, linear_form=None):
    r"""Computes the Kronecker Representation of a system of polynomials

    INPUT:

    * ``system`` -- A system of polynomials in ``d`` variables
    * ``vs`` -- Variables of the system
    * ``linear_form`` -- (Optional) A linear combination of the
      input variables that separates the critical point solutions

    OUTPUT:

    A polynomial ``P`` and ``d`` polynomials ``Q1, ..., Q_d`` such that
    ``z_i = Q_i(u)/P'(u)`` for ``u`` ranging over the roots of ``P``.

    Examples::
        sage: from sage_acsv.kronecker import kronecker
        sage: var('x,y')
        (x, y)
        sage: kronecker([x**3+y**3-10, y**2-2], [x,y], x+y)
        (u_^6 - 6*u_^4 - 20*u_^3 + 36*u_^2 - 120*u_ + 100,
         [60*u_^3 - 72*u_^2 + 360*u_ - 600, 12*u_^4 - 72*u_^2 + 240*u_])
    """
    R, u_ = PolynomialRing(QQ, "u_").objgen()
    R = PolynomialRing(QQ, len(vs) + 1, vs + [u_])
    system = [R(f) for f in system]
    vs = [R(v) for v in vs]
    u_ = R(u_)
    return _kronecker_representation(system, u_, vs, linear_form)
